# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
import dataclasses
from enum import auto, Enum
from functools import lru_cache
import numpy as np
import numpy.typing as npt
from queue import PriorityQueue
from typing import Generator, Optional, Union, Literal

from dyada.ancestrybranch import AncestryBranch
from dyada.descriptor import (
    RefinementDescriptor,
    get_regular_refined,
    hierarchical_to_box_index_mapping,
    int8_ndarray_from_iterable,
    find_uniqueness_violations,
)
from dyada.discretization import Discretization
from dyada.linearization import TrackToken
from dyada.mappings import correct_index_mapping, merge_mappings
from dyada.markers import (
    MarkerType,
    MarkersType,
    MarkersMapProxyType,
    get_next_largest_markered_index,
    get_defaultdict_for_markers,
)
from dyada.downsplit import apply_planned_downsplits


def is_lru_cached(func):
    while hasattr(func, "__wrapped__"):
        if hasattr(func, "cache_info"):
            return True
        func = func.__wrapped__
    return hasattr(func, "cache_info")


class PlannedAdaptiveRefinement:
    def __init__(self, discretization: Discretization):
        self._discretization = discretization
        # initialize planned refinement list and data structures used later
        self._planned_refinements: list[tuple[int, npt.NDArray[np.int8]]] = []
        self._planned_downsplits: list[tuple[int, ba.bitarray]] = []
        self._markers: MarkersType = get_defaultdict_for_markers(
            self._discretization.descriptor.get_num_dimensions()
        )
        self._upward_queue: PriorityQueue[tuple[int, int]] = PriorityQueue()

    def plan_coarsening(
        self,
        index: int,
        dimensions_to_coarsen: ba.bitarray,
    ) -> None:
        """plan to coarsen a node in the given directions (which merges its children);
        will raise error if this is not possible

        Args:
            index (int): hierarchical index of the node to coarsen
            dimensions_to_coarsen (ba.bitarray): 1 denotes if the node should be coarsened
        """
        num_dimensions = self._discretization.descriptor.get_num_dimensions()
        assert len(dimensions_to_coarsen) == num_dimensions
        dimensions_not_to_coarsen = ~dimensions_to_coarsen
        # children = self._discretization.descriptor.get_siblings(index + 1) #TODO add more checks
        parent = index

        # assert that the parent refinement has a 1 where coarsening is requested
        parent_refinement = self._discretization.descriptor[parent]
        assert (parent_refinement | dimensions_not_to_coarsen).count() == num_dimensions

        np_dimensions_to_coarsen = np.fromiter(
            dimensions_to_coarsen,
            dtype=np.int8,
            count=num_dimensions,
        )
        # then, store planned refinements, but the inverted one for the parent node
        self._planned_refinements.append((parent, -np_dimensions_to_coarsen))

    def plan_refinement(self, box_index: int, dimensions_to_refine=None) -> None:
        if dimensions_to_refine is None:
            dimensions_to_refine = (
                "1" * self._discretization.descriptor.get_num_dimensions()
            )
        # must be iterable, convert to np.array
        dimensions_to_refine = int8_ndarray_from_iterable(dimensions_to_refine)
        # get hierarchical index
        linear_index = self._discretization.descriptor.to_hierarchical_index(box_index)
        # store by linear index
        self._planned_refinements.append(
            (
                linear_index,
                dimensions_to_refine,
            )
        )

    def plan_downsplit(self, parent_index: int, dimensions_to_downsplit: ba.bitarray):
        """Move selected parent refinements one level down to all direct children.

        This operation intentionally de-normalizes the tree: selected split bits
        are removed from the parent and added to each current child.
        """
        descriptor = self._discretization.descriptor
        num_dimensions = descriptor.get_num_dimensions()
        if len(dimensions_to_downsplit) != num_dimensions:
            raise ValueError(
                "dimensions_to_downsplit length does not match discretization dimensionality"
            )
        if not (0 <= parent_index < len(descriptor)):
            raise IndexError("parent_index out of bounds")
        if dimensions_to_downsplit.count() == 0:
            return

        parent_refinement = descriptor[parent_index]
        if (dimensions_to_downsplit & ~parent_refinement).count() > 0:
            raise ValueError(
                "Can only downsplit dimensions that are currently refined at parent_index"
            )
        self._planned_downsplits.append((parent_index, dimensions_to_downsplit.copy()))

    def _apply_downsplits_direct(
        self, track_mapping: Literal["boxes", "patches"]
    ) -> tuple[Discretization, list[set[int]]]:
        if self._planned_refinements:
            raise NotImplementedError(
                "Mixing plan_downsplit with additional planned refinements is not supported yet"
            )
        mapping_result = apply_planned_downsplits(
            self._discretization, self._planned_downsplits, track_mapping
        )
        self._planned_downsplits = []
        return mapping_result

    def initialize_markers(self) -> None:
        # put initial markers
        for linear_index, dimensions_to_refine in self._planned_refinements:
            self._markers[linear_index] += int8_ndarray_from_iterable(
                dimensions_to_refine
            )
        self._planned_refinements = []  # clear the planned refinements

    def populate_queue(self) -> None:
        assert self._upward_queue.empty()
        self.initialize_markers()
        # put into the upward queue
        for linear_index in self._markers.keys():
            # obtain level sum to know the priority, highest level should come first
            level_sum = sum(
                self._discretization.descriptor.get_level(linear_index, False)
            )
            self._upward_queue.put((-level_sum, linear_index))

    def move_marker_to_parent(self, marker: MarkerType, sibling_indices) -> None:
        assert len(sibling_indices) > 1 and len(sibling_indices).bit_count() == 1
        sibling_indices = sorted(sibling_indices)
        # subtract from the current sibling markers
        for sibling in sibling_indices:
            self._markers[sibling] -= marker
            if np.all(self._markers[sibling] == np.zeros(marker.shape, dtype=np.int8)):
                self._markers.pop(sibling)

        # and add it to the parent's marker
        parent = sibling_indices[0] - 1
        self._markers[parent] += marker

    def move_marker_to_descendants(
        self, ancestor_index, marker: MarkerType, descendants_indices=None
    ):
        if np.all(marker == np.zeros(marker.shape, dtype=np.int8)):
            return
        if descendants_indices is None:
            # assume we want the direct children
            # TODO accept optional branch argument to reduce lookup time
            descendants_indices = self._discretization.descriptor.get_children(
                ancestor_index
            )
        assert (
            len(descendants_indices) > 1 and len(descendants_indices).bit_count() == 1
        )
        assert ancestor_index < min(descendants_indices)

        # subtract from the ancestor
        self._markers[ancestor_index] -= marker
        if np.all(
            self._markers[ancestor_index] == np.zeros(marker.shape, dtype=np.int8)
        ):
            self._markers.pop(ancestor_index)

        # and add to the descendants' markers
        for descendant in descendants_indices:
            self._markers[descendant] += marker

    def upwards_sweep(self) -> None:
        # traverse the tree from down (high level sums) to the coarser levels
        while not self._upward_queue.empty():
            level_sum, linear_index = self._upward_queue.get()

            if linear_index != 0:  # only continue upwards if not yet at the root
                # check if refinement can be moved up the branch (even partially);
                # this requires that all siblings are or would be refined
                siblings = self._discretization.descriptor.get_siblings(linear_index)
                all_siblings_refinements = [
                    int8_ndarray_from_iterable(
                        self._discretization.descriptor[sibling],
                    )
                    for sibling in siblings
                ]

                for i, sibling in enumerate(siblings):
                    if sibling in self._markers:
                        all_siblings_refinements[i] += self._markers[sibling]

                # check where the siblings are all refined
                possible_to_move_up = np.min(all_siblings_refinements, axis=0)
                assert np.all(possible_to_move_up >= 0)

                if np.any(possible_to_move_up > 0):
                    self.move_marker_to_parent(possible_to_move_up, siblings)

                    # put the parent into the upward queue
                    parent_level_sum = level_sum + (
                        len(all_siblings_refinements).bit_length() - 1
                    )
                    assert parent_level_sum > level_sum and parent_level_sum <= 0
                    parent = siblings[0] - 1
                    self._upward_queue.put((parent_level_sum, parent))

                # if any siblings were in the upward queue, remove them too
                for sibling in siblings:
                    if (level_sum, sibling) in self._upward_queue.queue:
                        self._upward_queue.queue.remove((level_sum, sibling))
            self._upward_queue.task_done()

    def downwards_sweep(self) -> None:
        descriptor = self._discretization.descriptor
        if len(self._markers) == 0:
            return

        def get_unresolvable_marker(
            current_refinement: ba.frozenbitarray, marker: MarkerType
        ) -> MarkerType:
            marker_to_push_down = marker.copy()

            # 1st case: negative but cannot be used to coarsen here
            marker_negative = marker_to_push_down < 0
            can_be_coarsened = int8_ndarray_from_iterable(
                current_refinement,
            )
            marker_to_push_down[marker_negative] += can_be_coarsened[marker_negative]

            # 2nd case: positive but cannot be used to refine here
            marker_positive = marker_to_push_down > 0
            can_be_refined = int8_ndarray_from_iterable(
                ~current_refinement,
            )
            marker_to_push_down[marker_positive] -= can_be_refined[marker_positive]
            return marker_to_push_down

        current_index = min(self._markers.keys())
        while True:
            if not descriptor.is_box(current_index):
                # check if (parts of) the marker need pushing down
                marker_to_push_down = get_unresolvable_marker(
                    descriptor[current_index], self._markers[current_index]
                )
                self.move_marker_to_descendants(current_index, marker_to_push_down)

            filtered_markers = {
                k: v for k, v in self._markers.items() if current_index < k
            }
            current_index = min(filtered_markers.keys(), default=-2)
            while current_index != -2 and not np.any(
                filtered_markers[current_index]
            ):  # remove zero markers
                self._markers.pop(current_index)
                filtered_markers.pop(current_index)
                current_index = min(filtered_markers.keys(), default=-2)
            if current_index == -2:
                # no more indices left in markers
                break

    @dataclasses.dataclass
    class Refinement:
        class Type(Enum):
            CopyOver = auto()
            TrackOnly = auto()
            ExpandLeaf = auto()

        type: "Type"
        old_index: int
        new_refinement: ba.bitarray | None = None
        marker_or_ancestor: MarkerType | int | None = None

    def _yield_missing_relationships(
        self,
        missing_mapping_items,
        map_tracking_tokens_to_new_indices: dict[TrackToken, int],
    ) -> Generator[Refinement, None, None]:
        missing_mappings: dict[int, set[int]] = {
            key: {map_tracking_tokens_to_new_indices[i] for i in indices}
            for key, indices in missing_mapping_items
        }
        for old_index, new_indices in missing_mappings.items():
            for new_index in new_indices:
                yield self.Refinement(
                    self.Refinement.Type.TrackOnly,
                    old_index,
                    None,
                    new_index,
                )

    def modified_branch_generator(
        self, starting_index: int, new_descriptor: RefinementDescriptor
    ) -> Generator[Refinement, None, None]:
        """Yields what every new entry in the new descriptor should be, and how to track

        Args:
            starting_index (int): where to start the branch
            new_descriptor (RefinementDescriptor): read-only access to the new descriptor

        Yields:
            Generator[Refinement, None, None]: sequence of refinement actions
        """
        descriptor = self._discretization.descriptor
        proxy_markers = MarkersMapProxyType(self._markers)
        abranch = AncestryBranch(self._discretization, starting_index, proxy_markers)
        map_tracking_tokens_to_new_indices: dict[TrackToken, int] = {}

        while True:
            current_new_index = len(new_descriptor)
            current_old_index, tracking_token, next_refinement, next_marker = (
                abranch.get_current_location_info()
            )
            map_tracking_tokens_to_new_indices[tracking_token] = current_new_index
            is_leaf = next_refinement == descriptor.d_zeros
            if is_leaf and np.min(next_marker) >= 0:
                # will actually be expanded
                yield self.Refinement(
                    self.Refinement.Type.ExpandLeaf,
                    current_old_index,
                    next_refinement,
                    next_marker,
                )
            else:
                yield self.Refinement(
                    self.Refinement.Type.CopyOver,
                    current_old_index,
                    next_refinement,
                )

            if is_leaf:
                # only on leaves, we advance the branch
                try:
                    abranch.advance()
                except (
                    AncestryBranch.WeAreDoneAndHereAreTheMissingRelationships
                ) as e:  # almost done!
                    yield from self._yield_missing_relationships(
                        e.missing_mapping.items(),
                        map_tracking_tokens_to_new_indices,
                    )
                    return
            else:
                # on non-leaves, we grow the branch
                abranch.grow(next_refinement)

    def track_indices(self, old_index: int, new_index: int) -> None:
        assert new_index > -1
        self._index_mapping[old_index].add(new_index)

    def extend_descriptor_and_track_indices(
        self,
        new_descriptor: RefinementDescriptor,
        old_range_to_extend: Union[int, tuple[int, int]],
        extension: ba.bitarray,
    ) -> None:
        previous_length = len(new_descriptor)
        new_descriptor._data.extend(extension)
        if isinstance(old_range_to_extend, int):
            for new_index in range(previous_length, len(new_descriptor)):
                self.track_indices(old_range_to_extend, new_index)
        else:
            for old_index, new_index in zip(
                range(old_range_to_extend[0], old_range_to_extend[1]),
                range(previous_length, len(new_descriptor)),
            ):
                self.track_indices(old_index, new_index)

    def add_refined_data(
        self, new_descriptor: RefinementDescriptor
    ) -> RefinementDescriptor:
        old_descriptor = self._discretization.descriptor
        last_extended_index = -1
        one_after_last_extended_index = 0
        while one_after_last_extended_index < len(old_descriptor):
            index_to_refine = get_next_largest_markered_index(
                self._markers, one_after_last_extended_index
            )
            if index_to_refine == -1:
                break

            self.extend_descriptor_and_track_indices(  # linearly copy up to marked
                new_descriptor,
                (one_after_last_extended_index, index_to_refine),
                old_descriptor[one_after_last_extended_index:index_to_refine],
            )
            # only refine where there are markers
            for requested_refinement in self.modified_branch_generator(
                index_to_refine, new_descriptor
            ):
                match requested_refinement:
                    case self.Refinement(
                        self.Refinement.Type.TrackOnly,
                        old_index,
                        None,
                        int() as new_index,
                    ):
                        # track only the index, namely the one of the previous (grand...)parent
                        self.track_indices(old_index, new_index)
                    case self.Refinement(
                        self.Refinement.Type.ExpandLeaf,
                        old_index,
                        ba.bitarray() as new_refinement,
                        np.ndarray() as new_marker,
                    ):
                        self.extend_descriptor_and_track_indices(
                            new_descriptor,
                            old_index,
                            get_regular_refined(new_marker),  # type: ignore
                        )
                    case self.Refinement(
                        self.Refinement.Type.CopyOver,
                        old_index,
                        ba.bitarray() as new_refinement,
                    ):
                        self.extend_descriptor_and_track_indices(
                            new_descriptor, old_index, new_refinement
                        )
                    case _:
                        raise RuntimeError("Logic error: should not be reached")
                last_extended_index = max(last_extended_index, old_index)
            one_after_last_extended_index = last_extended_index + 1

        # copy rest and return
        self.extend_descriptor_and_track_indices(
            new_descriptor,
            (one_after_last_extended_index, len(old_descriptor)),
            old_descriptor[one_after_last_extended_index : len(old_descriptor)],
        )
        return new_descriptor

    def create_new_discretization(
        self, track_mapping: Literal["boxes", "patches"] = "boxes"
    ) -> tuple[Discretization, list[set[int]]]:
        old_descriptor = self._discretization.descriptor
        self._index_mapping: list[set[int]] = [
            set() for _ in range(len(old_descriptor))
        ]

        # start generating the new descriptor
        new_descriptor = RefinementDescriptor(old_descriptor.get_num_dimensions())
        new_descriptor._data = ba.bitarray()

        # we are not changing the old descriptor, and greedily build the new one
        # so we can cache the box indices of both
        if not is_lru_cached(old_descriptor.to_box_index):
            old_descriptor.to_box_index = lru_cache(maxsize=None)(
                old_descriptor._to_box_index_recursive
            )
            old_descriptor._to_box_index_recursive = lru_cache(maxsize=None)(
                old_descriptor._to_box_index_recursive
            )
        new_descriptor.to_box_index = lru_cache(maxsize=None)(  # type: ignore
            new_descriptor._to_box_index_recursive
        )
        new_descriptor._to_box_index_recursive = lru_cache(maxsize=None)(  # type: ignore
            new_descriptor._to_box_index_recursive
        )

        new_descriptor = self.add_refined_data(new_descriptor)
        new_discretization = Discretization(
            self._discretization._linearization, new_descriptor
        )
        if track_mapping == "boxes":
            # transform the mapping to box indices
            self._index_mapping = hierarchical_to_box_index_mapping(
                self._index_mapping,
                old_descriptor,
                new_descriptor,
            )
        elif track_mapping == "patches":
            # may need to normalize in case refinement happened at non-leaves
            correct_index_mapping(
                self._index_mapping,
                self._discretization,
                new_discretization,
                self._markers,
            )
        else:
            raise ValueError(
                "track_mapping must be either 'boxes' or 'patches', got "
                + str(track_mapping)
            )
        return (
            new_discretization,
            self._index_mapping,
        )

    def apply_refinements(
        self,
        track_mapping: Literal["boxes", "patches"] = "boxes",
        sweep_mode: Literal["canonical", "as_planned"] = "canonical",
    ) -> tuple[Discretization, list[set[int]]]:
        assert self._upward_queue.empty()
        if sweep_mode == "canonical":
            assert len(self._markers) == 0
            self.populate_queue()
            self.upwards_sweep()
            self.downwards_sweep()
        elif sweep_mode == "as_planned":
            if self._planned_downsplits:
                return self._apply_downsplits_direct(track_mapping)
            self.initialize_markers()
        else:
            raise ValueError(
                "sweep_mode must be either 'canonical' or 'as_planned', got "
                + str(sweep_mode)
            )

        assert len(self._planned_refinements) == 0

        return self.create_new_discretization(track_mapping)


def apply_single_refinement(
    discretization: Discretization,
    box_index: int,
    dimensions_to_refine: Optional[ba.bitarray] = None,
    track_mapping: Literal["boxes", "patches"] = "boxes",
) -> tuple[Discretization, list[set[int]]]:
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(box_index, dimensions_to_refine)
    return p.apply_refinements(track_mapping=track_mapping)


def normalize_discretization(
    discretization: Discretization,
    track_mapping: Literal["boxes", "patches"] = "patches",
    max_normalization_rounds: int = 2**31 - 1,
) -> tuple[Discretization, list[set[int]], int]:
    """
    Normalize the discretization so that it fulfills the uniqueness condition
    and we get a normalized omnitree.
    """
    normalization_rounds = 0
    # find the tuples of indices where the uniqueness condition is violated
    violations = find_uniqueness_violations(discretization.descriptor)
    mapping: list[set[int]] = []
    while len(violations) > 0 and normalization_rounds < max_normalization_rounds:
        normalization_rounds += 1
        p = PlannedAdaptiveRefinement(discretization)
        # remove these violations, by putting markers and executing create_new_descriptor
        for violation in violations:
            # find the dimension(s) of the violation
            sorted_violation = sorted(violation)
            dimensions_to_shift = ~ba.bitarray(
                discretization.descriptor[sorted_violation[0]]
            )
            for i in sorted_violation[1:]:
                dimensions_to_shift &= discretization.descriptor[i]
            assert dimensions_to_shift.count() > 0
            dimensions_to_shift_array = int8_ndarray_from_iterable(
                dimensions_to_shift,
            )
            p._markers[sorted_violation[0]] += dimensions_to_shift_array
            for i in sorted_violation[1:]:
                p._markers[i] -= dimensions_to_shift_array

        # apply the refinements
        new_discretization, new_mapping = p.create_new_discretization(
            track_mapping=track_mapping
        )
        discretization = new_discretization
        mapping = merge_mappings(mapping, new_mapping)
        violations = find_uniqueness_violations(discretization.descriptor)

    return discretization, mapping, normalization_rounds



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
import pytest

from dyada.linearization import (
    MortonOrderLinearization,
    TrackToken,
    binary_or_none_generator,
    get_initial_coarsening_stack,
    get_initial_coarsen_refine_stack,
    indices_to_bitmask,
)


def test_get_position_morton_order():
    lin = MortonOrderLinearization()

    level_increment = ba.bitarray("1")
    position = lin.get_binary_position_from_index([0], [level_increment])
    assert position == ba.bitarray("0")
    position = lin.get_binary_position_from_index([1], [level_increment])
    assert position == ba.bitarray("1")
    with pytest.raises(IndexError):
        lin.get_binary_position_from_index([2], [level_increment])

    level_increment = ba.bitarray("11")
    position = lin.get_binary_position_from_index([0], [level_increment])
    assert position == ba.bitarray("00")
    position = lin.get_binary_position_from_index([1], [level_increment])
    assert position == ba.bitarray("10")
    position = lin.get_binary_position_from_index([2], [level_increment])
    assert position == ba.bitarray("01")
    position = lin.get_binary_position_from_index([3], [level_increment])
    assert position == ba.bitarray("11")
    with pytest.raises(IndexError):
        lin.get_binary_position_from_index([4], [level_increment])

    level_increment = ba.bitarray("01")
    position = lin.get_binary_position_from_index([0], [level_increment])
    assert position == ba.bitarray("00")
    position = lin.get_binary_position_from_index([1], [level_increment])
    assert position == ba.bitarray("01")
    with pytest.raises(IndexError):
        lin.get_binary_position_from_index([2], [level_increment])

    level_increment = ba.bitarray("10")
    position = lin.get_binary_position_from_index([0], [level_increment])
    assert position == ba.bitarray("00")
    position = lin.get_binary_position_from_index([1], [level_increment])
    assert position == ba.bitarray("10")
    with pytest.raises(IndexError):
        lin.get_binary_position_from_index([2], [level_increment])

    level_increment = ba.bitarray("111")
    position = lin.get_binary_position_from_index([0], [level_increment])
    assert position == ba.bitarray("000")
    position = lin.get_binary_position_from_index([1], [level_increment])
    assert position == ba.bitarray("100")
    position = lin.get_binary_position_from_index([2], [level_increment])
    assert position == ba.bitarray("010")
    position = lin.get_binary_position_from_index([3], [level_increment])
    assert position == ba.bitarray("110")
    position = lin.get_binary_position_from_index([7], [level_increment])
    assert position == ba.bitarray("111")
    with pytest.raises(IndexError):
        lin.get_binary_position_from_index([8], [level_increment])

    level_increment = ba.bitarray("110")
    position = lin.get_binary_position_from_index([0], [level_increment])
    assert position == ba.bitarray("000")
    position = lin.get_binary_position_from_index([1], [level_increment])
    assert position == ba.bitarray("100")
    position = lin.get_binary_position_from_index([2], [level_increment])
    assert position == ba.bitarray("010")
    position = lin.get_binary_position_from_index([3], [level_increment])
    assert position == ba.bitarray("110")
    with pytest.raises(IndexError):
        lin.get_binary_position_from_index([4], [level_increment])


def test_get_index_morton_order():
    lin = MortonOrderLinearization()

    level_increment = ba.bitarray("1")
    index = lin.get_index_from_binary_position(ba.bitarray("0"), [], [level_increment])
    assert index == 0
    index = lin.get_index_from_binary_position(ba.bitarray("1"), [], [level_increment])
    assert index == 1
    with pytest.raises(AssertionError):
        lin.get_index_from_binary_position(ba.bitarray("10"), [], [level_increment])

    level_increment = ba.bitarray("11")
    position = lin.get_index_from_binary_position(
        ba.bitarray("00"), [], [level_increment]
    )
    assert position == 0
    position = lin.get_index_from_binary_position(
        ba.bitarray("10"), [], [level_increment]
    )
    assert position == 1
    position = lin.get_index_from_binary_position(
        ba.bitarray("01"), [], [level_increment]
    )
    assert position == 2
    position = lin.get_index_from_binary_position(
        ba.bitarray("11"), [], [level_increment]
    )
    assert position == 3

    level_increment = ba.bitarray("01")
    position = lin.get_index_from_binary_position(
        ba.bitarray("00"), [], [level_increment]
    )
    assert position == 0
    position = lin.get_index_from_binary_position(
        ba.bitarray("01"), [], [level_increment]
    )
    assert position == 1
    with pytest.raises(AssertionError):
        lin.get_index_from_binary_position(ba.bitarray("10"), [], [level_increment])
    with pytest.raises(AssertionError):
        lin.get_index_from_binary_position(ba.bitarray("11"), [], [level_increment])

    level_increment = ba.bitarray("10")
    position = lin.get_index_from_binary_position(
        ba.bitarray("00"), [], [level_increment]
    )
    assert position == 0
    position = lin.get_index_from_binary_position(
        ba.bitarray("10"), [], [level_increment]
    )
    assert position == 1

    level_increment = ba.bitarray("111")
    position = lin.get_index_from_binary_position(
        ba.bitarray("000"), [], [level_increment]
    )
    assert position == 0
    position = lin.get_index_from_binary_position(
        ba.bitarray("100"), [], [level_increment]
    )
    assert position == 1
    position = lin.get_index_from_binary_position(
        ba.bitarray("010"), [], [level_increment]
    )
    assert position == 2
    position = lin.get_index_from_binary_position(
        ba.bitarray("110"), [], [level_increment]
    )
    assert position == 3
    position = lin.get_index_from_binary_position(
        ba.bitarray("111"), [], [level_increment]
    )
    assert position == 7


def test_binary_or_none_generator():
    indices = {0, 2}
    N = 4
    generated = list(binary_or_none_generator(indices, N))
    expected = [
        (0, None, 0, None),
        (0, None, 0, None),
        (0, None, 0, None),
        (0, None, 0, None),
        (1, None, 0, None),
        (1, None, 0, None),
        (1, None, 0, None),
        (1, None, 0, None),
        (0, None, 1, None),
        (0, None, 1, None),
        (0, None, 1, None),
        (0, None, 1, None),
        (1, None, 1, None),
        (1, None, 1, None),
        (1, None, 1, None),
        (1, None, 1, None),
    ]
    assert generated == expected


def _pop_all_positions(tracker):
    """Pop all positions from a tracker, returning them in pop order."""
    positions = []
    while len(tracker) > 0:
        pos, _ = tracker.pop()
        positions.append(pos)
    return positions


def test_empty_coarsening_stack_initialization():
    fba = ba.frozenbitarray
    tracker = get_initial_coarsening_stack(
        current_parent_refinement=fba("111"),
        dimensions_to_coarsen=fba("000"),
    )
    assert tracker.separated_mask == fba("000")
    assert tracker.unresolved_mask is None
    expected_pop_order = [
        fba("000"),
        fba("100"),
        fba("010"),
        fba("110"),
        fba("001"),
        fba("101"),
        fba("011"),
        fba("111"),
    ]
    assert _pop_all_positions(tracker) == expected_pop_order


def test_all_coarsening_stack_initialization():
    fba = ba.frozenbitarray
    tracker = get_initial_coarsening_stack(
        current_parent_refinement=fba("111"),
        dimensions_to_coarsen=fba("111"),
    )
    assert tracker.separated_mask == fba("111")
    expected_pop_order = [
        fba("000"),
        fba("100"),
        fba("010"),
        fba("110"),
        fba("001"),
        fba("101"),
        fba("011"),
        fba("111"),
    ]
    assert _pop_all_positions(tracker) == expected_pop_order


def test_coarsening_stack_2d():
    tracker = get_initial_coarsening_stack(
        current_parent_refinement=ba.frozenbitarray("11"),
        dimensions_to_coarsen=indices_to_bitmask((0,), 2),
    )
    assert tracker.separated_mask == ba.frozenbitarray("10")
    expected_pop_order = [
        ba.frozenbitarray("00"),
        ba.frozenbitarray("01"),
        ba.frozenbitarray("10"),
        ba.frozenbitarray("11"),
    ]
    assert _pop_all_positions(tracker) == expected_pop_order


def test_coarsening_stack_3d():
    fba = ba.frozenbitarray
    tracker = get_initial_coarsening_stack(
        current_parent_refinement=fba("111"),
        dimensions_to_coarsen=indices_to_bitmask((0, 1), 3),
    )
    assert tracker.separated_mask == fba("110")

    # Pop first item and register its group
    first_pos, first_same = tracker.pop()
    assert first_pos == fba("000")
    assert first_same is None

    tracker.register_group(first_pos, TrackToken(42))

    # Remaining pops should resolve groups via the dict
    expected = [
        (fba("001"), None),
        (fba("100"), TrackToken(42)),
        (fba("101"), None),
        (fba("010"), TrackToken(42)),
        (fba("011"), None),
        (fba("110"), TrackToken(42)),
        (fba("111"), None),
    ]
    for expected_pos, expected_same in expected:
        pos, same = tracker.pop()
        assert pos == expected_pos
        assert same == expected_same


def test_coarsen_refine_stack_cannot_coarsen_2d():
    fba = ba.frozenbitarray
    tracker = get_initial_coarsen_refine_stack(
        current_parent_refinement=fba("01"),
        dimensions_to_coarsen=fba("00"),
        dimensions_to_refine=fba("00"),
        dimensions_cannot_coarsen=fba("10"),
    )
    assert tracker.separated_mask == fba("00")
    assert tracker.unresolved_mask == fba("10")
    expected_pop_order = [fba("00"), fba("01"), fba("10"), fba("11")]
    assert _pop_all_positions(tracker) == expected_pop_order


def test_coarsen_refine_stack_3d():
    fba = ba.frozenbitarray
    tracker = get_initial_coarsen_refine_stack(
        current_parent_refinement=ba.frozenbitarray("011"),
        dimensions_to_coarsen=indices_to_bitmask((2,), 3),
        dimensions_to_refine=indices_to_bitmask((0,), 3),
    )
    assert tracker.separated_mask == fba("001")

    # Pop first and register group
    first_pos, first_same = tracker.pop()
    assert first_pos == fba("000")
    assert first_same is None
    tracker.register_group(first_pos, TrackToken(99))

    # Only pos "001" shares remaining_positions "00" with first_pos "000"
    expected = [
        (fba("100"), None),
        (fba("010"), None),
        (fba("110"), None),
        (fba("001"), TrackToken(99)),
        (fba("101"), None),
        (fba("011"), None),
        (fba("111"), None),
    ]
    for expected_pos, expected_same in expected:
        pos, same = tracker.pop()
        assert pos == expected_pos
        assert same == expected_same



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
import pytest
from os.path import abspath
from dyada.drawing import discretization_to_2d_ascii

from dyada.descriptor import RefinementDescriptor
from dyada.discretization import Discretization
from dyada.linearization import MortonOrderLinearization
from dyada.refinement import (
    PlannedAdaptiveRefinement,
)


def test_coarsen_simplest_2d():
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(2, ba.bitarray("10 00 00")),
    )
    p = PlannedAdaptiveRefinement(discretization)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(4, 2))
        == """\
_____
| | |
|_|_|"""
    )
    p.plan_coarsening(0, ba.bitarray("10"))
    new_discretization, _ = p.apply_refinements()
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("00")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(4, 2))
        == """\
_____
|   |
|___|"""
    )

    # aaand transposed
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(2, ba.bitarray("01 00 00")),
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("01"))
    new_discretization, _ = p.apply_refinements()
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("00")


def test_coarsen_partly_2d():
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(2, ba.bitarray("11 00 00 00 00")),
    )
    assert (
        discretization_to_2d_ascii(discretization, resolution=(4, 2))
        == """\
_____
|_|_|
|_|_|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("10"))
    new_discretization, index_mapping = p.apply_refinements(track_mapping="patches")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(4, 2))
        == """\
_____
|___|
|___|"""
    )
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("01 00 00")
    expected_index_mapping = {0: {0}, 1: {1}, 2: {1}, 3: {2}, 4: {2}}
    assert index_mapping == [
        expected_index_mapping[i] for i in range(len(expected_index_mapping))
    ]

    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("01"))
    new_discretization, index_mapping = p.apply_refinements(track_mapping="patches")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(4, 2))
        == """\
_____
| | |
|_|_|"""
    )
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("10 00 00")

    expected_index_mapping = {0: {0}, 1: {1}, 2: {2}, 3: {1}, 4: {2}}
    assert index_mapping == [
        expected_index_mapping[i] for i in range(len(expected_index_mapping))
    ]


def test_coarsen_octree_first():
    for dimensionality in range(1, 5):
        desc_initial = RefinementDescriptor(dimensionality, [2] * dimensionality)
        discretization_initial = Discretization(
            MortonOrderLinearization(), desc_initial
        )
        # coarsen first parent
        all_coarsening = ba.bitarray("1" * dimensionality)
        coarsen_first_oct_plan = PlannedAdaptiveRefinement(discretization_initial)
        coarsen_first_oct_plan.plan_coarsening(1, all_coarsening)
        first_coarsened_discretization, coarsen_first_oct_mapping = (
            coarsen_first_oct_plan.apply_refinements(track_mapping="patches")
        )
        first_coarsened_descriptor = first_coarsened_discretization.descriptor
        assert first_coarsened_descriptor[1].count() == 0
        remaining_length = len(first_coarsened_descriptor) - 2
        assert (
            first_coarsened_descriptor[-remaining_length:]
            == desc_initial[-remaining_length:]
        )
        assert coarsen_first_oct_mapping[0] == {0}
        for i in range(1, 2**dimensionality + 2):
            assert coarsen_first_oct_mapping[i] == {1}
        for i in range(2**dimensionality + 2, len(desc_initial)):
            assert coarsen_first_oct_mapping[i] == {i - 2**dimensionality}


def test_coarsen_octree_all_2d():
    dimensionality = 2
    desc_initial = RefinementDescriptor(dimensionality, [2] * dimensionality)
    discretization_initial = Discretization(MortonOrderLinearization(), desc_initial)
    assert (
        discretization_to_2d_ascii(discretization_initial, resolution=(8, 4))
        == """\
_________
|_|_|_|_|
|_|_|_|_|
|_|_|_|_|
|_|_|_|_|"""
    )

    # coarsen all parents
    all_coarsening = ba.bitarray("1" * dimensionality)
    coarsen_plan = PlannedAdaptiveRefinement(discretization_initial)
    coarsen_plan.plan_coarsening(1, all_coarsening)
    coarsen_plan.plan_coarsening(6, all_coarsening)
    coarsen_plan.plan_coarsening(11, all_coarsening)
    coarsen_plan.plan_coarsening(16, all_coarsening)
    coarsened_discretization, coarsen_mapping = coarsen_plan.apply_refinements(
        track_mapping="patches"
    )
    coarsened_descriptor = coarsened_discretization.descriptor
    assert coarsened_descriptor == RefinementDescriptor(
        dimensionality, [1] * dimensionality
    )
    assert (
        discretization_to_2d_ascii(coarsened_discretization, resolution=(4, 2))
        == """\
_____
|_|_|
|_|_|"""
    )
    expected_coarsen_mapping = [{0}, *[{1}] * 5, *[{2}] * 5, *[{3}] * 5, *[{4}] * 5]
    assert coarsen_mapping == expected_coarsen_mapping


def test_flip_2d():
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(2, ba.bitarray("01 00 00")),
    )
    assert (
        discretization_to_2d_ascii(discretization, resolution=(4, 2))
        == """\
_____
|___|
|___|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "10")
    p.plan_refinement(1, "10")
    p.plan_coarsening(0, ba.bitarray("01"))
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("10 00 00")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(4, 2))
        == """\
_____
| | |
|_|_|"""
    )

    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 2},
        2: {0, 1, 2},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]


def test_flip_3d():
    # needs the info from test_coarsen_refine_stack_3d to be correct
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(3, ba.bitarray("011 000 000 000 000")),
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "100")
    p.plan_refinement(1, "100")
    p.plan_refinement(2, "100")
    p.plan_refinement(3, "100")
    p.plan_coarsening(0, ba.bitarray("001"))
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("110 000 000 000 000")
    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 2},
        2: {0, 3, 4},
        3: {0, 1, 2},
        4: {0, 3, 4},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]


def test_coarsen_right_half_2d():
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(2, ba.bitarray("11 00 00 00 00")),
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("01"))  # <- hierarchical index
    p.plan_refinement(0, "01")  # <- box/leaf index
    # p.plan_refinement(2, "01") # necessary?
    assert (
        discretization_to_2d_ascii(discretization, resolution=(4, 2))
        == """\
_____
|_|_|
|_|_|"""
    )
    new_discretization, index_mapping = p.apply_refinements(track_mapping="patches")
    new_descriptor = new_discretization.descriptor
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(4, 2))
        == """\
_____
|_| |
|_|_|"""
    )
    assert new_descriptor._data == ba.bitarray("10 01 00 00 00")
    expected_index_mapping = {
        0: {0},
        1: {2},
        2: {4},
        3: {3},
        4: {4},
    }
    assert index_mapping == [
        expected_index_mapping[i] for i in range(len(expected_index_mapping))
    ]


def test_coarsen_nested_2d():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 00 01 00 00 01 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == """\
_________
|___|___|
|___|___|
|   |   |
|___|___|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("10"))  # <- hierarchical index
    new_discretization, index_mapping = p.apply_refinements(track_mapping="patches")
    new_descriptor = new_discretization.descriptor
    assert new_descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 00 01 00 00")
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(8, 4))
        == """\
_________
|_______|
|_______|
|       |
|_______|"""
    )
    expected_index_mapping = [{0}, {1}, {1}, {2}, {3}, {4}, {2}, {3}, {4}]
    assert index_mapping == expected_index_mapping


def test_coarsen_nested_3d_whole():
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "111 000 000 010 000 000 010 000 000 000 000 010 000 000 010 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    expected_index_mapping = {
        "patches": [
            *[{0}, {1}, {1}, {2}, {3}, {4}, {2}, {3}, {4}],
            *[{1}, {1}, {2}, {3}, {4}, {2}, {3}, {4}],
        ],
        "boxes": [{0}, {0}, {1}, {2}, {1}, {2}, {0}, {0}, {1}, {2}, {1}, {2}],
    }
    for refinement in ["boxes", "patches"]:
        p = PlannedAdaptiveRefinement(discretization)
        p.plan_coarsening(0, ba.bitarray("101"))  # <- hierarchical index
        new_discretization, index_mapping = p.apply_refinements(
            track_mapping=refinement
        )
        new_descriptor = new_discretization.descriptor
        assert new_descriptor == RefinementDescriptor.from_binary(
            3, ba.bitarray("010 000 010 000 000")
        )
        assert index_mapping == expected_index_mapping[refinement]


def test_coarsen_nested_3d_half():
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "111 000 000 010 000 000 010 000 000 000 000 010 000 000 010 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    expected_index_mapping = {
        "patches": [
            *[{0}, {1}, {1}, {2}, {3}, {4}, {2}, {3}],
            *[{4}, {5}, {5}, {6}, {7}, {8}, {6}, {7}, {8}],
        ],
        "boxes": [{0}, {0}, {1}, {2}, {1}, {2}, {3}, {3}, {4}, {5}, {4}, {5}],
    }
    for refinement in ["boxes", "patches"]:
        p = PlannedAdaptiveRefinement(discretization)
        p.plan_coarsening(0, ba.bitarray("100"))  # <- hierarchical index
        new_discretization, index_mapping = p.apply_refinements(
            track_mapping=refinement
        )
        new_descriptor = new_discretization.descriptor
        assert new_descriptor == RefinementDescriptor.from_binary(
            3, ba.bitarray("011 000 010 000 000 000 010 000 000")
        )
        assert index_mapping == expected_index_mapping[refinement]


def test_coarsen_double_nested_2d():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 00 01 00 01 00 00 01 00 01 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 8))
        == """\
_________
|___|___|
|___|___|
|   |   |
|___|___|
|   |   |
|   |   |
|   |   |
|___|___|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("10"))  # <- hierarchical index
    new_discretization, index_mapping = p.apply_refinements(track_mapping="patches")
    new_descriptor = new_discretization.descriptor
    assert new_descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 00 01 00 01 00 00")
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(8, 8))
        == """\
_________
|_______|
|_______|
|       |
|_______|
|       |
|       |
|       |
|_______|"""
    )
    expected_index_mapping = [
        *[{0}, {1}, {1}],
        *[{2}, {3}, {4}, {5}, {6}, {2}, {3}, {4}, {5}, {6}],
    ]
    assert index_mapping == expected_index_mapping


def test_coarsen_refine_nested_2d():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 00 01 00 00 01 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == """\
_________
|___|___|
|___|___|
|   |   |
|___|___|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_coarsening(0, ba.bitarray("10"))
    p.plan_refinement(0, ba.bitarray("10"))
    p.plan_refinement(2, ba.bitarray("10"))
    new_discretization, index_mapping = p.apply_refinements(
        track_mapping="patches", sweep_mode="as_planned"
    )
    new_descriptor = new_discretization.descriptor
    assert new_descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 00 00 01 10 00 00 00")
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(8, 4))
        == """\
_________
|_______|
|___|___|
|   |   |
|___|___|"""
    )
    expected_index_mapping = {
        0: {0},
        1: {2},
        2: {3},
        3: {4},
        4: {6},
        5: {8},
        6: {4},
        7: {7},
        8: {8},
    }
    assert index_mapping == [
        expected_index_mapping[i] for i in range(len(expected_index_mapping))
    ]
    down_plan = PlannedAdaptiveRefinement(discretization)
    down_plan.plan_downsplit(0, ba.bitarray("10"))
    down_discretization, down_mapping = down_plan.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    down_descriptor = down_discretization.descriptor
    assert down_descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 00 00 11 00 00 00 00")
    )
    down_mapping_expected = {
        0: {0, 1, 4},
        1: {2},
        2: {3},
        3: {4},
        4: {5},
        5: {7},
        6: {4},
        7: {6},
        8: {8},
    }
    assert down_mapping == [
        down_mapping_expected[i] for i in range(len(down_mapping_expected))
    ]


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections import Counter
import pytest
import bitarray as ba
import numpy as np
from os.path import abspath

from dyada.descriptor import (
    RefinementDescriptor,
    validate_descriptor,
    hierarchical_to_box_index_mapping,
)
from dyada.discretization import (
    coordinates_from_box_index,
    coordinates_from_index,
)
from dyada.refinement import (
    Discretization,
    PlannedAdaptiveRefinement,
    apply_single_refinement,
)
from dyada.linearization import MortonOrderLinearization, single_bit_set_gen


def helper_check_box_mapping(
    index_mapping: list[set[int]],
    old_discretization: Discretization,
    new_discretization: Discretization,
    tested_refinement=None,
):
    old_descriptor = old_discretization.descriptor
    new_descriptor = new_discretization.descriptor
    assert len(index_mapping) == old_descriptor.get_num_boxes()
    assert set() not in index_mapping
    count_new_indices: Counter = Counter()
    for value in index_mapping:
        count_new_indices.update(value)
    # check that each is there once
    for e in count_new_indices.most_common():
        assert e[1] == 1
    assert sorted(count_new_indices.elements()) == list(
        range(new_descriptor.get_num_boxes())
    )
    for mapped_from_index, mapped_to_indices in enumerate(index_mapping):
        old_coordinates = coordinates_from_box_index(
            old_discretization, mapped_from_index
        )
        if len(mapped_to_indices) == 1:
            (mapped_to_index,) = mapped_to_indices
            new_coordinates = coordinates_from_box_index(
                new_discretization, mapped_to_index
            )
            # make sure the coordinates are correct
            assert old_coordinates == new_coordinates
        else:
            # assert that the new boxes cover the old box
            # new_coordinates_list = [
            #     coordinates_from_index(new_discretization, new_index, is_box_index=True)
            #     for new_index in mapped_to_indices
            # ]
            # # needs recursive line-sweep #TODO
            pass


def helper_check_mapping(
    index_mapping: list[set[int]],
    old_discretization: Discretization,
    new_discretization: Discretization,
    mapping_indices_are_boxes=True,
    tested_refinement=None,
) -> None:
    if mapping_indices_are_boxes:
        return helper_check_box_mapping(
            index_mapping, old_discretization, new_discretization, tested_refinement
        )
    else:
        # get box mapping from index mapping
        box_mapping = hierarchical_to_box_index_mapping(
            index_mapping,
            old_discretization.descriptor,
            new_discretization.descriptor,
        )
        helper_check_box_mapping(box_mapping, old_discretization, new_discretization)

    old_descriptor = old_discretization.descriptor
    new_descriptor = new_discretization.descriptor
    count_new_indices: Counter = Counter()
    for value in index_mapping:
        count_new_indices.update(value)
    assert len(index_mapping) == len(old_descriptor)
    assert set() not in index_mapping
    assert set(count_new_indices.elements()) == set(range(len(new_descriptor)))
    for b, mapped_to_indices in enumerate(index_mapping):
        old_coordinates = coordinates_from_index(
            old_discretization, b, mapping_indices_are_boxes
        )
        # the smallest of the mapped_to_indices' interval should cover the old interval
        most_senior_mapped_to = min(mapped_to_indices)
        new_coordinates = coordinates_from_index(
            new_discretization,
            most_senior_mapped_to,
            mapping_indices_are_boxes,
        )
        assert np.all(new_coordinates.lower_bound <= old_coordinates.lower_bound)
        assert np.all(old_coordinates.upper_bound <= new_coordinates.upper_bound)
        if len(mapped_to_indices) > 1:
            # and this should not be true for the second
            second_most_senior_mapped_to = sorted(mapped_to_indices)[1]
            new_coordinates = coordinates_from_index(
                new_discretization,
                second_most_senior_mapped_to,
                mapping_indices_are_boxes,
            )
            assert not (
                np.all(new_coordinates.lower_bound <= old_coordinates.lower_bound)
                and np.all(old_coordinates.upper_bound <= new_coordinates.upper_bound)
            )


def test_refine_4d():
    descriptor = RefinementDescriptor(4, 0)
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, ba.bitarray("0010"))
    discretization, _ = p.apply_refinements()
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, ba.bitarray("1101"))
    discretization, _ = p.apply_refinements()
    p = PlannedAdaptiveRefinement(discretization)
    last_box_index = discretization.descriptor.get_num_boxes() - 1
    p.plan_refinement(last_box_index, ba.bitarray("0001"))

    new_discretization, index_mapping = p.apply_refinements(track_mapping="boxes")

    assert new_discretization.descriptor._data == ba.bitarray(
        "0011110000000000000000000000110000000000000000000000"
    )
    helper_check_mapping(index_mapping, discretization, new_discretization)


def test_refine_random():
    for d in range(1, 5):
        discretization = Discretization(
            MortonOrderLinearization(), RefinementDescriptor(d, 0)
        )
        for round_number in range(3):
            descriptor = discretization.descriptor
            p = PlannedAdaptiveRefinement(discretization)
            round_refinements = []
            for _ in range(2 ** (round_number + d // 2)):
                random_box = np.random.randint(0, descriptor.get_num_boxes())
                random_refinement = ba.bitarray(
                    (np.random.randint(0, 2) for _ in range(d))
                )
                p.plan_refinement(random_box, random_refinement)
                round_refinements.append((random_box, random_refinement))

            new_discretization, index_mapping = p.apply_refinements(
                track_mapping="patches"
            )
            box_mapping = hierarchical_to_box_index_mapping(
                index_mapping, descriptor, new_discretization.descriptor
            )

            # the "=" may happen if only zeros are chosen
            assert len(new_discretization.descriptor) >= len(descriptor)
            assert validate_descriptor(new_discretization.descriptor)
            helper_check_mapping(
                index_mapping,
                discretization,
                new_discretization,
                mapping_indices_are_boxes=False,
                tested_refinement=round_refinements,
            )
            helper_check_mapping(
                box_mapping,
                discretization,
                new_discretization,
                mapping_indices_are_boxes=True,
                tested_refinement=round_refinements,
            )
            discretization = new_discretization


def test_refine_random_increments():
    for d in range(1, 5):
        descriptor = RefinementDescriptor(d, 0)
        assert descriptor.get_num_boxes() == 1
        possible_increments = list(single_bit_set_gen(d))
        for round_number in range(10 + 4 ^ d):
            random_box = np.random.randint(0, descriptor.get_num_boxes())
            random_refinement = possible_increments[np.random.randint(0, d)]

            discretization = Discretization(MortonOrderLinearization(), descriptor)
            new_discretization, index_mapping = apply_single_refinement(
                discretization,
                random_box,
                random_refinement,
            )
            new_descriptor = new_discretization.descriptor
            assert new_descriptor.get_num_boxes() == descriptor.get_num_boxes() + 1
            assert validate_descriptor(new_descriptor)
            try:
                helper_check_mapping(index_mapping, discretization, new_discretization)
                for b in range(descriptor.get_num_boxes()):
                    assert len(index_mapping[b]) == 1 or (
                        b == random_box and len(index_mapping[b]) == 2
                    )
            except Exception as e:
                print(
                    f"failed for round {round_number} with box {random_box} and refinement {random_refinement}"
                )
                print(f"old descriptor: {descriptor}")
                print(f"new descriptor: {new_descriptor}")
                raise e
            descriptor = new_descriptor


def test_refinement_invalid_track_mapping():
    descriptor = RefinementDescriptor(2, 0)
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, ba.bitarray("11"))
    with pytest.raises(ValueError, match="track_mapping"):
        p.apply_refinements(track_mapping="invalid")


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
from itertools import permutations
import matplotlib.pyplot as plt
import numpy as np
import os
import pytest

from dyada.coordinates import (
    get_coordinates_from_level_index,
    interval_from_sequences,
    level_index_from_sequence,
)
from dyada.descriptor import RefinementDescriptor, validate_descriptor
from dyada.discretization import (
    Discretization,
)
from dyada.drawing import (
    plot_boxes_2d,
    plot_all_boxes_2d,
    boxes_to_2d_ascii,
    plot_all_boxes_3d,
)
from dyada.drawing_tikz import (
    latex_write_and_compile,
    plot_tree_tikz,
    plot_descriptor_tikz,
    plot_location_stack_tikz,
)
from dyada.linearization import MortonOrderLinearization
from dyada.refinement import (
    PlannedAdaptiveRefinement,
)
from dyada.structure import depends_on_optional, module_is_available


def test_depends_on_optional():
    @depends_on_optional("nonexistent_module")
    def should_not_run():
        pass

    with pytest.raises(ImportError):
        should_not_run()


def test_no_latex_error():
    with pytest.warns(UserWarning):
        latex_write_and_compile("test", "test_latex_no_error.tex")
    os.remove("test_latex_no_error.tex")
    # check that there are no files with "test_latex_no_error" in the current directory
    files = [f for f in os.listdir(".") if os.path.isfile(f)]
    for f in files:
        assert "test_latex_no_error" not in f


# todo consider comparing images: https://github.com/matplotlib/pytest-mpl
def test_plot_boxes_2d():
    with plt.ion():  # turns off blocking figures for test
        for backend in ["matplotlib", "tikz", "ascii"]:
            level_index = level_index_from_sequence([0, 0], [0, 0])
            coordinates = get_coordinates_from_level_index(level_index)
            plot_boxes_2d([coordinates], labels=[str(level_index)], backend=backend)
            level_index = level_index_from_sequence([0, 1, 0], [0, 0, 0])
            level_index2 = level_index_from_sequence([0, 1, 0], [0, 1, 0])
            coordinates = get_coordinates_from_level_index(level_index)
            coordinates2 = get_coordinates_from_level_index(level_index2)
            plot_boxes_2d(
                [coordinates, coordinates2],
                labels=["0", "1"],
                alpha=0.2,
                backend=backend,
            )
            level_index = level_index_from_sequence(
                [5, 4, 3, 2, 1, 0], [31, 15, 7, 3, 1, 0]
            )
            coordinates = get_coordinates_from_level_index(level_index)
            plot_boxes_2d([coordinates], projection=[3, 4], backend=backend)

    with pytest.raises(AssertionError):
        plot_boxes_2d([coordinates], projection=[2, 3, 4])

    with pytest.raises(ValueError):
        plot_boxes_2d([coordinates], backend="unknown")


def test_plot_boxes_2d_from_descriptor():
    descriptor = RefinementDescriptor(4, [0, 1, 2, 3])
    r = Discretization(MortonOrderLinearization(), descriptor)
    with plt.ion():  # turns off blocking figures for test
        for backend in ["matplotlib", "tikz", "ascii"]:
            # try all combinations of projections
            for projection in permutations(range(4), 2):
                # the transparency should give murky colors for the lower projections
                # (because many boxes will be stacked on top of each other)
                # and nicer colors for the higher ones
                plot_all_boxes_2d(
                    r,
                    projection=list(projection),
                    alpha=0.3,
                    filename="2d_transparent",
                    backend=backend,
                    colors="orange",
                )
                plot_all_boxes_2d(
                    r, projection=list(projection), labels="boxes", backend=backend
                )


def test_plot_complex_2d_with_stack():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("10 01 00 00 10 01 00 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)

    plot_all_boxes_2d(
        discretization,
        backend="tikz",
        filename="complex_2d_square",
    )
    plot_descriptor_tikz(descriptor, filename="complex_2d_desc")
    plot_tree_tikz(descriptor, filename="complex_2d_tree")
    plot_location_stack_tikz(discretization, filename="complex_2d_location_stack")
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(2, "01")
    p.plan_refinement(4, "02")
    non_normalized_discretization, _ = p.apply_refinements()
    non_normalized_descriptor = non_normalized_discretization.descriptor
    assert non_normalized_descriptor._data == ba.bitarray(
        "11 00 10 01 00 00 01 00 00 00 10 00 01 00 00"
    )
    plot_all_boxes_2d(
        non_normalized_discretization,
        backend="tikz",
        filename="complex_2d_square_after",
    )
    plot_descriptor_tikz(non_normalized_descriptor, filename="complex_2d_desc_after")
    plot_tree_tikz(non_normalized_descriptor, filename="complex_2d_tree_after")
    plot_location_stack_tikz(
        non_normalized_discretization, filename="complex_2d_location_stack_after"
    )


def test_boxes_to_2d_ascii():
    omnitree_cells = [
        interval_from_sequences((0.0, 0.0), (0.25, 1.0)),
        interval_from_sequences((0.25, 0.0), (0.5, 0.5)),
        interval_from_sequences((0.25, 0.5), (0.5, 1.0)),
        interval_from_sequences((0.5, 0.0), (1.0, 1.0)),
    ]
    assert (
        boxes_to_2d_ascii(omnitree_cells, projection=[0, 1], resolution=(32, 16))
        == """\
_________________________________
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |_______|               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|       |       |               |
|_______|_______|_______________|"""
    )
    assert (
        boxes_to_2d_ascii(omnitree_cells, projection=[1, 0], resolution=(16, 8))
        == """\
_________________
|               |
|               |
|               |
|_______________|
|       |       |
|_______|_______|
|               |
|_______________|"""
    )
    assert (
        boxes_to_2d_ascii(omnitree_cells, projection=[0, 1], resolution=(8, 4))
        == """\
_________
| | |   |
| |_|   |
| | |   |
|_|_|___|"""
    )
    with pytest.warns(UserWarning):
        ascii_4_2 = boxes_to_2d_ascii(
            omnitree_cells, projection=[0, 1], resolution=(4, 2)
        )
    assert (
        ascii_4_2
        == """\
_____
||| |
|||_|"""
    )
    with pytest.warns(UserWarning):
        ascii_2_1 = boxes_to_2d_ascii(
            omnitree_cells, projection=[0, 1], resolution=(2, 1)
        )
    assert (
        ascii_2_1
        == """\
___
|||"""
    )


def test_draw_simplest_grandchild_split_tikz():
    # cf. test_refine_simplest_grandchild_split
    descriptor_1 = RefinementDescriptor.from_binary(
        2, ba.bitarray("10 01 00 00 10 00 00")
    )
    descriptor_2 = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 10 00 00 00 10 00 00")
    )
    # transpose refinements of the two above
    descriptor_3 = RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 00 00 01 00 00")
    )
    descriptor_4 = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 00 01 00 00 01 00 00")
    )

    plot_tree_tikz(descriptor_1, filename="grandchild_split_before")
    plot_all_boxes_2d(
        Discretization(MortonOrderLinearization(), descriptor_1),
        backend="tikz",
        filename="grandchild_split_before_2d",
        labels="patches",
        connect_centers=True,
    )
    plot_all_boxes_2d(
        Discretization(MortonOrderLinearization(), descriptor_2),
        backend="tikz",
        filename="grandchild_split_after_2d",
        labels=None,
        connect_centers=True,
    )
    plot_all_boxes_2d(
        Discretization(MortonOrderLinearization(), descriptor_3),
        backend="tikz",
        filename="grandchild_split_before_2d_transpose",
        labels="patches",
        projection=[1, 0],
        connect_centers=True,
    )  # should be the same as the first
    plot_all_boxes_2d(
        Discretization(MortonOrderLinearization(), descriptor_4),
        backend="tikz",
        filename="grandchild_split_after_2d_transpose",
        labels=None,
        projection=[1, 0],
        connect_centers=True,
    )  # should be different from the second


def test_plot_boxes_3d_from_descriptor():
    descriptor = RefinementDescriptor(3, [1, 0, 1])
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(3, ba.bitarray("101"))
    p.plan_refinement(1, ba.bitarray("001"))
    p.plan_refinement(2, ba.bitarray("010"))
    discretization, _ = p.apply_refinements()
    validate_descriptor(discretization.descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(3, ba.bitarray("100"))
    p.plan_refinement(7, ba.bitarray("010"))
    discretization, _ = p.apply_refinements()
    validate_descriptor(discretization.descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(9, ba.bitarray("101"))
    discretization, _ = p.apply_refinements()
    validate_descriptor(discretization.descriptor)
    new_descriptor = discretization.descriptor
    backends = ["tikz", "obj"]
    backends.append("matplotlib") if module_is_available("matplotlib") else None
    backends.append("opengl") if module_is_available("OpenGL") else None
    (
        backends.append("aaaaargh") if module_is_available("aaaaargh") else None
    )  # should not raise
    for backend in backends:
        if backend == "tikz":
            plot_all_boxes_3d(
                discretization,
                labels="boxes",
                draw_options="fill opacity=0.1",
                backend=backend,
            )
            plot_tree_tikz(
                new_descriptor,
                labels=["ö" + str(a) for a in np.arange(len(new_descriptor))],
            )
            plot_descriptor_tikz(new_descriptor)
        else:
            with plt.ion():  # turns off blocking figures for test
                plot_all_boxes_3d(
                    discretization, labels="boxes", alpha=0.1, backend=backend
                )
        plot_all_boxes_3d(
            discretization,
            wireframe=True,
            filename="test_filename_wireframe",
            backend=backend,
        )


def test_plot_octree_3d_from_descriptor():
    descriptor = RefinementDescriptor(3, [1, 1, 1])
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(4, ba.bitarray("111"))
    p.plan_refinement(7, ba.bitarray("111"))
    new_discretization, _ = p.apply_refinements()
    validate_descriptor(new_discretization.descriptor)
    p = PlannedAdaptiveRefinement(new_discretization)
    p.plan_refinement(
        new_discretization.descriptor.get_num_boxes() - 2, ba.bitarray("111")
    )
    new_discretization, _ = p.apply_refinements()
    new_descriptor = new_discretization.descriptor
    backends = ["tikz", "obj"]
    backends.append("matplotlib") if module_is_available("matplotlib") else None
    backends.append("opengl") if module_is_available("OpenGL") else None
    for backend in backends:
        if backend == "matplotlib" or backend == "opengl":
            plot_all_boxes_3d(
                new_discretization,
                labels="boxes",
                filename="octree",
                alpha=0.1,
                backend=backend,
                colors="red",
                linewidth=3.0,
            )
        else:
            plot_all_boxes_3d(
                new_discretization,
                labels="boxes",
                filename="octree",
                draw_options="fill opacity=0.1",
                backend=backend,
            )
    plot_all_boxes_3d(
        new_discretization, wireframe=True, filename="octree", labels=None
    )
    plot_tree_tikz(new_descriptor, filename="octree_tree")
    with pytest.raises(ValueError):
        plot_all_boxes_3d(new_discretization, filename="octree_tree", backend="unknown")



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
import pytest
import numpy as np
from os.path import abspath

from dyada.ancestrybranch import AncestryBranch
from dyada.descriptor import RefinementDescriptor
from dyada.discretization import Discretization
from dyada.drawing import discretization_to_2d_ascii
from dyada.linearization import MortonOrderLinearization, TrackToken
from dyada.refinement import PlannedAdaptiveRefinement, get_defaultdict_for_markers


def advance_or_grow(
    ancestrybranch: AncestryBranch, next_refinement: ba.frozenbitarray
) -> AncestryBranch:
    is_leaf = next_refinement.count() == 0
    if is_leaf:
        ancestrybranch.advance()
    else:
        # on non-leaves, we grow the branch
        ancestrybranch.grow(next_refinement)
    return ancestrybranch


def test_ancestry_branch_2d_6():
    # inspired by test_refine_2d_6
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("01 00 10 10 00 00 10 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == """\
_________
| | | | |
|_|_|_|_|
|       |
|_______|"""
    )
    markers = get_defaultdict_for_markers(
        discretization.descriptor.get_num_dimensions()
    )
    markers[0] = np.array([1, 0], dtype=np.int8)
    markers[6] = np.array([-1, 0], dtype=np.int8)

    ancestrybranch = AncestryBranch(discretization, starting_index=0, markers=markers)
    # step through ancestry branch
    expected_return_values = [
        (0, TrackToken(0), ba.frozenbitarray("11"), np.array([1, 0], dtype=np.int8)),
        (1, TrackToken(1), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (1, TrackToken(2), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (2, TrackToken(3), ba.frozenbitarray("10"), np.array([0, 0], dtype=np.int8)),
        (3, TrackToken(4), ba.frozenbitarray("10"), np.array([0, 0], dtype=np.int8)),
        (4, TrackToken(5), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (4, TrackToken(6), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (3, TrackToken(7), ba.frozenbitarray("10"), np.array([0, 0], dtype=np.int8)),
        (5, TrackToken(8), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (5, TrackToken(9), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (2, TrackToken(10), ba.frozenbitarray("10"), np.array([0, 0], dtype=np.int8)),
        (7, TrackToken(11), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (8, TrackToken(12), ba.frozenbitarray("00"), np.array([0, 0], dtype=np.int8)),
        (None),
    ]
    for expected in expected_return_values:
        # new index given by track_token
        current_old_index, track_token, next_refinement, next_marker = (
            ancestrybranch.get_current_location_info()
        )
        assert (current_old_index, track_token, next_refinement) == expected[0:-1]
        assert np.array_equal(next_marker, expected[-1])
        try:
            ancestrybranch = advance_or_grow(ancestrybranch, next_refinement)
        except AncestryBranch.WeAreDoneAndHereAreTheMissingRelationships as e:
            assert track_token == TrackToken(12)  # make sure it happens at end
            assert e.missing_mapping == {
                1: {TrackToken(0)},
                2: {TrackToken(0)},
                3: {TrackToken(0), TrackToken(3)},
                4: {TrackToken(0), TrackToken(3), TrackToken(4)},
                5: {TrackToken(0), TrackToken(3), TrackToken(7)},
                6: {TrackToken(10), TrackToken(11), TrackToken(12)},
            }
            break


def test_ancestrybranch_3d_5():
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("001 010 100 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)

    markers = get_defaultdict_for_markers(
        discretization.descriptor.get_num_dimensions()
    )
    markers[0] = np.array([1, 1, 0], dtype=np.int8)
    markers[1] = np.array([0, -1, 0], dtype=np.int8)
    markers[2] = np.array([-1, 0, 0], dtype=np.int8)
    ancestrybranch = AncestryBranch(discretization, starting_index=0, markers=markers)

    TT = TrackToken  # type alias for brevity
    leaf_no_marker = (ba.frozenbitarray("000"), np.array([0, 0, 0], dtype=np.int8))
    # step through ancestry branch
    expected_return_values = [
        (
            0,
            TT(0),
            ba.frozenbitarray("111"),
            np.array([1, 1, 0], dtype=np.int8),
        ),
        (3, TT(1), *leaf_no_marker),
        (4, TT(2), *leaf_no_marker),
        (5, TT(3), *leaf_no_marker),
        (5, TT(4), *leaf_no_marker),
        (6, TT(5), *leaf_no_marker),
        (6, TT(6), *leaf_no_marker),
        (6, TT(7), *leaf_no_marker),
        (6, TT(8), *leaf_no_marker),
        (None),
    ]
    for expected in expected_return_values:
        current_old_index, track_token, next_refinement, next_marker = (
            ancestrybranch.get_current_location_info()
        )
        assert (current_old_index, track_token, next_refinement) == expected[0:-1]
        assert np.array_equal(next_marker, expected[-1])
        try:
            ancestrybranch = advance_or_grow(ancestrybranch, next_refinement)
        except AncestryBranch.WeAreDoneAndHereAreTheMissingRelationships as e:
            assert track_token == TT(8)
            assert e.missing_mapping == {
                1: {TT(0), TT(1), TT(2), TT(3), TT(4)},
                2: {TT(0), TT(1), TT(2)},
                5: {TT(0)},
                6: {TT(0)},
            }
            break


def test_ancestrybranch_coarsen_nested_2d():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 00 01 00 00 01 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)

    markers = get_defaultdict_for_markers(
        discretization.descriptor.get_num_dimensions()
    )
    markers[0] = np.array([-1, 0], dtype=np.int8)
    ancestrybranch = AncestryBranch(discretization, starting_index=0, markers=markers)

    TT = TrackToken
    bba = ba.bitarray
    fba = ba.frozenbitarray
    no_marker = np.array([0, 0], dtype=np.int8)
    leaf = fba("00")
    # step through ancestry branch
    expected_return_values = [
        (0, TT(0), fba("01"), np.array([-1, 0], dtype=np.int8)),
        (1, TT(1), leaf, no_marker),
        (3, TT(2), fba("01"), no_marker),
        (4, TT(3), bba("00"), no_marker),
        (5, TT(4), leaf, no_marker),
        (None),
    ]
    for expected in expected_return_values:
        current_old_index, track_token, next_refinement, next_marker = (
            ancestrybranch.get_current_location_info()
        )
        assert (current_old_index, track_token, next_refinement) == expected[0:-1]
        assert np.array_equal(next_marker, expected[-1])
        try:
            ancestrybranch = advance_or_grow(ancestrybranch, next_refinement)
        except AncestryBranch.WeAreDoneAndHereAreTheMissingRelationships as e:
            assert track_token == TT(4)
            assert e.missing_mapping == {
                1: {TT(0)},
                2: {TT(1)},
                3: {TT(0)},
                4: {TT(0), TT(2)},
                5: {TT(0), TT(2)},
                6: {TT(2)},
                7: {TT(3)},
                8: {TT(4)},
            }
            break


def test_modified_branch_generator_2d_6():
    descriptor = RefinementDescriptor.from_binary(
        2,
        ba.bitarray("01 00 10 10 00 00 10 00 00"),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == """\
_________
| | | | |
|_|_|_|_|
|       |
|_______|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p._markers[0] = np.array([1, 0], dtype=np.int8)
    p._markers[6] = np.array([-1, 0], dtype=np.int8)
    new_descriptor = RefinementDescriptor(2)
    new_descriptor._data = ba.bitarray()
    p._index_mapping = [set() for _ in range(len(discretization.descriptor))]
    # call like in add_refined_data
    generator = p.modified_branch_generator(
        starting_index=0, new_descriptor=new_descriptor
    )
    zeroth_refinement = next(generator)
    assert zeroth_refinement == p.Refinement(
        p.Refinement.Type.CopyOver, 0, ba.frozenbitarray("11"), None
    )
    p.extend_descriptor_and_track_indices(
        new_descriptor, zeroth_refinement.old_index, zeroth_refinement.new_refinement
    )
    assert p._index_mapping == [{0}, *[set()] * 8]
    first_refinement = next(generator)
    assert first_refinement.type == p.Refinement.Type.ExpandLeaf
    assert first_refinement.old_index == 1
    assert first_refinement.new_refinement == ba.bitarray("00")
    assert np.array_equal(
        first_refinement.marker_or_ancestor, np.array([0, 0], dtype=np.int8)
    )
    p.extend_descriptor_and_track_indices(
        new_descriptor,
        first_refinement.old_index,
        ba.bitarray("00"),
    )
    assert p._index_mapping == [{0}, {1}, *[set()] * 7]

    second_refinement = next(generator)
    assert second_refinement.type == p.Refinement.Type.ExpandLeaf
    assert second_refinement.old_index == 1
    assert second_refinement.new_refinement == ba.bitarray("00")
    assert np.array_equal(
        second_refinement.marker_or_ancestor, np.array([0, 0], dtype=np.int8)
    )
    p.extend_descriptor_and_track_indices(
        new_descriptor,
        first_refinement.old_index,
        ba.bitarray("00"),
    )

    third_refinement = next(generator)
    assert third_refinement == p.Refinement(
        p.Refinement.Type.CopyOver, 2, ba.bitarray("10"), None
    )
    p.extend_descriptor_and_track_indices(
        new_descriptor,
        third_refinement.old_index,
        third_refinement.new_refinement,
    )

    fourth_refinement = next(generator)
    assert fourth_refinement == p.Refinement(
        p.Refinement.Type.CopyOver, 3, ba.bitarray("10"), None
    )
    p.extend_descriptor_and_track_indices(
        new_descriptor,
        fourth_refinement.old_index,
        fourth_refinement.new_refinement,
    )
    assert new_descriptor._data == ba.bitarray("11 00 00 10 10")


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
import bitarray.util
import dataclasses
from functools import lru_cache
import numpy as np
import numpy.typing as npt
from typing import NamedTuple, TypeAlias, Sequence

# todo move LocationCode etc into coordinates?
from dyada.linearization import LocationCode


class DyadaTooFineError(ValueError):
    pass


@dataclasses.dataclass
class LevelIndex:
    d_level: npt.NDArray[np.int8]
    d_index: npt.NDArray[np.int64]

    def __iter__(self):
        """make iterable, mainly to allow unpacking in assignments"""
        return iter(dataclasses.astuple(self))

    def __eq__(self, other: object):
        if not isinstance(other, LevelIndex):
            return NotImplemented
        return np.all(self.d_level == other.d_level) and np.all(
            self.d_index == other.d_index
        )


def level_index_from_sequence(
    d_level: Sequence[int], d_index: Sequence[int]
) -> LevelIndex:
    if max(d_level) > 62:
        raise DyadaTooFineError(
            f"Level too large, maximum 1d level is 62, \
              got l={d_level}, i={d_index}"
        )
    return LevelIndex(
        np.asarray(d_level, dtype=np.int8),
        np.asarray(d_index, dtype=np.int64),
    )


Coordinate: TypeAlias = npt.NDArray[np.float64]


def coordinate_from_sequence(c: Sequence[float]) -> Coordinate:
    return np.asarray(c, dtype=np.float64)


class CoordinateInterval(NamedTuple):
    lower_bound: Coordinate
    upper_bound: Coordinate

    def __eq__(self, other: object):
        if not isinstance(other, CoordinateInterval):
            return NotImplemented
        return np.all(self.lower_bound == other.lower_bound) and np.all(
            self.upper_bound == other.upper_bound
        )


def interval_from_sequences(
    lower_bound: Sequence[float], upper_bound: Sequence[float]
) -> CoordinateInterval:
    return CoordinateInterval(
        coordinate_from_sequence(lower_bound),
        coordinate_from_sequence(upper_bound),
    )


def get_coordinates_from_level_index(level_index: LevelIndex) -> CoordinateInterval:
    num_dimensions = len(level_index.d_level)
    assert num_dimensions == len(level_index.d_index)
    if any(level_index.d_level > 62):
        raise DyadaTooFineError(
            f"Level too large, maximum 1d level is 62, \
              got l={level_index.d_level}, i={level_index.d_index}"
        )
    if (
        any(level_index.d_level < 0)
        or any(level_index.d_index < 0)
        or any(level_index.d_index >= 2 ** np.array(level_index.d_level, dtype=int))
    ):
        error_string = "Invalid level index: {}".format(level_index)
        if any(level_index.d_index >= 2 ** np.array(level_index.d_level, dtype=int)):
            error_string += " (index {} too large, should be < {})".format(
                level_index.d_index, 2 ** np.array(level_index.d_level, dtype=int)
            )
        raise ValueError(error_string)

    def get_d_array(x):
        return np.fromiter(
            x,
            dtype=np.float64,
            count=num_dimensions,
        )

    @lru_cache(maxsize=None)  # cache the function to avoid recomputing
    def get_neg_power_of_two(level):
        return 2.0 ** -float(level)

    return CoordinateInterval(
        get_d_array(
            (
                get_neg_power_of_two(level) * i
                for level, i in zip(level_index.d_level, level_index.d_index)
            )
        ),
        get_d_array(
            (
                get_neg_power_of_two(level) * (i + 1)
                for level, i in zip(level_index.d_level, level_index.d_index)
            )
        ),
    )


def location_code_from_coordinate(coordinate: Coordinate) -> LocationCode:
    return LocationCode(location_code_from_float(co) for co in coordinate)


def level_index_from_location_code(location_code: list[ba.bitarray]) -> LevelIndex:
    d_level = []
    d_index = []
    for bits in location_code:
        level = len(bits)
        if level == 0:
            index = 0
        else:
            index = bitarray.util.ba2int(bits)
        d_level.append(level)
        d_index.append(index)
    return level_index_from_sequence(d_level, d_index)


def location_code_from_level_index(level_index: LevelIndex) -> list[ba.bitarray]:
    location_code = []
    for d_level, d_index in zip(level_index.d_level, level_index.d_index):
        if d_level == 0:
            bits = ba.bitarray("")  # level 0 has empty location code
        else:
            bits = bitarray.util.int2ba(int(d_index), length=int(d_level))
        location_code.append(bits)
    return location_code


def float_parts_bitarray(value):
    dtype = value.dtype.type if hasattr(value, "dtype") else np.float64
    f = dtype(value)
    if dtype == np.float64:
        raw_bits = f.view(np.uint64)
        exp_bits = 11
        mant_bits = 52
    elif dtype == np.float32:
        raw_bits = f.view(np.uint32)
        exp_bits = 8
        mant_bits = 23
    else:
        raise ValueError("Only float32 and float64 are supported.")

    total_bits = exp_bits + mant_bits + 1
    bits = ba.bitarray(f"{raw_bits:0{total_bits}b}")

    sign_bit = bits[0]
    exponent = bits[1 : 1 + exp_bits]
    mantissa = bits[1 + exp_bits :]

    return sign_bit, exponent, mantissa


def location_code_from_float(value):
    if value < 0.0 or value > 1.0:
        raise ValueError("Value must be in the range [0.0, 1.0]")
    if value == 1.0:
        return ba.bitarray("1" * 23)
    sign_bit, exponent, mantissa = float_parts_bitarray(value + 1)
    assert sign_bit == 0
    # exponent represented as 2-complement
    assert bitarray.util.ba2int(exponent) == 2 ** (len(exponent) - 1) - 1
    return mantissa


def bitarray_startswith(
    long_bitarray: ba.bitarray, potentially_shorter_bitarray: ba.bitarray
) -> bool:
    return (
        long_bitarray[: len(potentially_shorter_bitarray)]
        == potentially_shorter_bitarray
    )



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import bitarray as ba
import numpy as np
import pytest
from random import randint

from dyada.coordinates import (
    DyadaTooFineError,
    LevelIndex,
    bitarray_startswith,
    level_index_from_sequence,
    interval_from_sequences,
    get_coordinates_from_level_index,
    float_parts_bitarray,
    location_code_from_float,
    location_code_from_coordinate,
    location_code_from_level_index,
    level_index_from_location_code,
)


def test_get_coordinates_from_level_index():
    level_index = level_index_from_sequence([0, 0], [0, 0])
    assert get_coordinates_from_level_index(level_index) == interval_from_sequences(
        [0.0, 0.0], [1.0, 1.0]
    )
    with pytest.raises(ValueError):
        get_coordinates_from_level_index(level_index_from_sequence([0, 0], [1, 0]))
    level_index = level_index_from_sequence([0, 1, 0], [0, 1, 0])
    assert get_coordinates_from_level_index(level_index) == interval_from_sequences(
        [0.0, 0.5, 0.0], [1.0, 1.0, 1.0]
    )
    level_index = level_index_from_sequence([5, 4, 3, 2, 1, 0], [31, 15, 7, 3, 1, 0])
    expected_lower = [0.96875, 0.9375, 0.875, 0.75, 0.5, 0.0]
    assert get_coordinates_from_level_index(level_index) == interval_from_sequences(
        expected_lower, [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    )
    level_index = level_index_from_sequence([0, 1, 2, 3], [0, 0, 0, 0])
    assert get_coordinates_from_level_index(level_index) == interval_from_sequences(
        [0.0, 0.0, 0.0, 0.0], [1.0, 0.5, 0.25, 0.125]
    )
    # make sure that if too large, we get a DyadaTooFineError
    with pytest.raises(DyadaTooFineError):
        level_index_from_sequence([0, 63, 2, 3], [0, 0, 0, 0])
    barely_ok_level_index = level_index_from_sequence([0, 62, 2, 3], [0, 0, 0, 0])
    barely_ok_level_index.d_level[1] = 63  # this should be too fine
    with pytest.raises(DyadaTooFineError):
        get_coordinates_from_level_index(barely_ok_level_index)

    # some randomization
    for _ in range(100):
        num_dimensions = randint(1, 10)
        level = [randint(0, 12) for _ in range(num_dimensions)]
        index = [randint(0, 2**l_d - 1) for l_d in level]
        level_index = LevelIndex(
            np.asarray(level, dtype=np.uint8), np.asarray(index, dtype=np.uint16)
        )  # use of uint is intentional!

        interval = get_coordinates_from_level_index(level_index)
        assert np.all(interval.lower_bound < interval.upper_bound)
        assert np.all(interval.lower_bound >= 0.0)
        assert np.all(interval.upper_bound <= 1.0)


def test_float_parts_bitarray():
    # test the float_parts_bitarray function
    # 1.0 should be 0b
    _, _, mantissa_bits = float_parts_bitarray(1.0)
    assert mantissa_bits.count() == 0
    # but putting 1.0 in the location code should give us 1111...
    mantissa_bits = location_code_from_float(1.0)
    assert mantissa_bits.count() == len(mantissa_bits)

    with pytest.raises(ValueError):
        float_parts_bitarray(np.longdouble(1.0))
    with pytest.raises(ValueError):
        float_parts_bitarray(np.half(1.0))


def test_location_code_from_coordinate():
    coordinate = np.array([0.0, 0.25, 0.5, 0.75])
    location_code = location_code_from_coordinate(coordinate)
    expected_codes = [
        ba.bitarray(""),  # 0.0
        ba.bitarray("01"),  # 0.25
        ba.bitarray("1"),  # 0.5
        ba.bitarray("11"),  # 0.75
    ]
    for code, expected in zip(location_code, expected_codes):
        assert bitarray_startswith(code, expected)


def test_conversions_random_data():
    assert len(ba.bitarray("0")) == 1
    for _ in range(100):
        num_dimensions = randint(1, 6)
        level = [randint(0, 10) for _ in range(num_dimensions)]
        index = [randint(0, 2**l_d - 1) for l_d in level]
        level_index = level_index_from_sequence(level, index)
        interval = get_coordinates_from_level_index(level_index)
        coordinate_from_li = interval.lower_bound

        location_code_from_c = location_code_from_coordinate(coordinate_from_li)
        location_code_from_li = location_code_from_level_index(level_index)
        for from_c, from_li in zip(location_code_from_c, location_code_from_li):
            assert bitarray_startswith(from_c, from_li)
        # coordinate_from_location_code = ...#TODO
        # assert np.allclose(coordinate_from_li, coordinate_from_location_code)

        level_index_from_lc = level_index_from_location_code(location_code_from_li)
        assert level_index == level_index_from_lc



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Literal

import bitarray as ba

from dyada.descriptor import (
    RefinementDescriptor,
    get_num_children_from_refinement,
    hierarchical_to_box_index_mapping,
)
from dyada.discretization import Discretization
from dyada.linearization import (
    get_initial_child_grouping,
)


def _merged_planned_downsplits(
    planned_downsplits: list[tuple[int, ba.bitarray]],
) -> dict[int, ba.bitarray]:
    merged: dict[int, ba.bitarray] = {}
    for parent_index, dimensions_to_downsplit in planned_downsplits:
        existing = merged.get(parent_index)
        if existing is None:
            merged[parent_index] = dimensions_to_downsplit.copy()
        else:
            merged[parent_index] |= dimensions_to_downsplit
    return merged


def _can_absorb(
    members: list[tuple[ba.frozenbitarray, int, int]],
    dims_to_split: ba.bitarray,
    descriptor: RefinementDescriptor,
) -> bool:
    """Check if grouped children can be absorbed into a single merged node."""
    old_refs = [descriptor[m[1]] for m in members]
    return (
        len(set(ba.frozenbitarray(r) for r in old_refs)) == 1
        and not (old_refs[0] & dims_to_split).any()
        and old_refs[0].count() > 0
    )


def apply_planned_downsplits(
    discretization: Discretization,
    planned_downsplits: list[tuple[int, ba.bitarray]],
    track_mapping: Literal["boxes", "patches"],
) -> tuple[Discretization, list[set[int]]]:
    """Apply all planned downsplits in a single forward pass over the descriptor.

    Uses the ChildGroupingTracker to group children of each downsplit node, then
    emits intermediate nodes and re-ordered children into the target descriptor.
    """
    descriptor = discretization.descriptor
    nd = descriptor.get_num_dimensions()
    linearization = discretization._linearization

    merged_downsplits = _merged_planned_downsplits(planned_downsplits)
    if not merged_downsplits:
        return discretization, [{i} for i in range(len(descriptor))]

    # Single forward DFS pass: walk old descriptor, emit new descriptor.
    new_data = ba.bitarray()
    mapping: list[set[int]] = [set() for _ in range(len(descriptor))]

    def _track(old_idx: int, new_idx: int) -> None:
        mapping[old_idx].add(new_idx)

    def _emit(ref_bits: ba.bitarray | ba.frozenbitarray) -> int:
        """Emit one node into the new descriptor, return its new index."""
        new_idx = len(new_data) // nd
        new_data.extend(ref_bits)
        return new_idx

    def _copy_range(old_start: int, old_end: int) -> None:
        """Copy a contiguous range of nodes from old to new descriptor."""
        new_start = len(new_data) // nd
        new_data.extend(descriptor._data[old_start * nd : old_end * nd])
        for offset in range(old_end - old_start):
            mapping[old_start + offset].add(new_start + offset)

    def _walk(old_start: int, size: int) -> None:
        """Copy or process a subtree, handling nested downsplits."""
        old_end = old_start + size
        i = old_start
        while i < old_end:
            if i in merged_downsplits:
                i = _process_downsplit(i)
            else:
                run_end = i + 1
                while run_end < old_end and run_end not in merged_downsplits:
                    run_end += 1
                _copy_range(i, run_end)
                i = run_end

    def _emit_absorbed(
        members: list[tuple[ba.frozenbitarray, int, int]],
        parent_old_idx: int,
        dims_to_split: ba.bitarray,
    ) -> None:
        """Emit a merged node absorbing the intermediate and its children.
        The children's refinement is combined with the pushed-down dims.
        Grandchildren are interleaved in merged order.
        """
        child_ref = descriptor[members[0][1]]
        merged_ref = ba.bitarray(child_ref) | dims_to_split
        new_merged_idx = _emit(merged_ref)
        _track(parent_old_idx, new_merged_idx)
        for _, old_start, _ in members:
            _track(old_start, new_merged_idx)

        # Collect grandchildren keyed by full position under merged_ref
        gc_entries: dict[ba.frozenbitarray, tuple[int, int]] = {}
        for local_pos, old_start, _ in members:
            old_child_ref = descriptor[old_start]
            for gc_idx, (gc_start, gc_end) in enumerate(
                descriptor.get_child_ranges(old_start)
            ):
                gc_pos = linearization.get_binary_position_from_index(
                    (gc_idx,), (old_child_ref,)
                )
                full_pos = ba.frozenbitarray(
                    (local_pos & dims_to_split) | (gc_pos & old_child_ref)
                )
                gc_entries[full_pos] = (gc_start, gc_end - gc_start)

        # Emit in merged order
        num_merged_gc = get_num_children_from_refinement(merged_ref)
        for m in range(num_merged_gc):
            m_pos = ba.frozenbitarray(
                linearization.get_binary_position_from_index((m,), (merged_ref,))
            )
            gc_old, gc_size = gc_entries[m_pos]
            _walk(gc_old, gc_size)

    def _emit_intermediate(
        members: list[tuple[ba.frozenbitarray, int, int]],
        parent_old_idx: int,
        dims_to_split: ba.bitarray,
    ) -> None:
        """Emit an intermediate node followed by each child subtree."""
        _track(parent_old_idx, _emit(ba.bitarray(dims_to_split)))
        for _, old_start, sz in members:
            _walk(old_start, sz)

    def _process_downsplit(old_idx: int) -> int:
        """Process a downsplit node, returning the old index after its subtree."""
        dims_to_split = merged_downsplits[old_idx]
        parent_ref = ba.bitarray(descriptor[old_idx])

        remaining_ref = parent_ref & ~dims_to_split

        if remaining_ref.count() == 0:
            raise ValueError(
                f"Downsplit at node {old_idx}: cannot push down all refined dims"
            )

        # Emit the parent with reduced ref
        _track(old_idx, _emit(remaining_ref))

        # Map each child's local position to (old_start, subtree_size)
        child_ranges = descriptor.get_child_ranges(old_idx)
        pos_to_info: dict[ba.frozenbitarray, tuple[int, int]] = {}
        for child_local_idx, (start, end) in enumerate(child_ranges):
            pos = ba.frozenbitarray(
                linearization.get_binary_position_from_index(
                    (child_local_idx,), (parent_ref,)
                )
            )
            pos_to_info[pos] = (start, end - start)
        subtree_end = child_ranges[-1][1]

        # Group children by remaining-position bits (dims NOT being pushed down).
        # Pass remaining_ref as sort_dimensions so the tracker sorts by remaining
        # bits first — making same-group members adjacent in pop order.
        tracker = get_initial_child_grouping(
            ba.frozenbitarray(parent_ref),
            ba.frozenbitarray(remaining_ref),
            linearization=linearization,
        )

        current_members: list[tuple[ba.frozenbitarray, int, int]] = []

        def _flush_group() -> None:
            if not current_members:
                return
            if _can_absorb(current_members, dims_to_split, descriptor):
                _emit_absorbed(current_members, old_idx, dims_to_split)
            else:
                _emit_intermediate(current_members, old_idx, dims_to_split)

        current_key: ba.frozenbitarray | None = None
        while len(tracker) > 0:
            local_pos, _ = tracker.pop()
            old_start, size = pos_to_info[local_pos]
            remaining_key = ba.frozenbitarray(local_pos & remaining_ref)
            if remaining_key != current_key:
                _flush_group()
                current_members = []
                current_key = remaining_key
            current_members.append((local_pos, old_start, size))
        _flush_group()

        return subtree_end

    # Walk the full descriptor
    _walk(0, len(descriptor))

    new_descriptor = RefinementDescriptor(nd)
    new_descriptor._data = new_data
    new_discretization = Discretization(linearization, new_descriptor)

    if track_mapping == "patches":
        return new_discretization, mapping
    elif track_mapping == "boxes":
        return new_discretization, hierarchical_to_box_index_mapping(
            mapping, descriptor, new_descriptor
        )
    else:
        raise ValueError(
            "track_mapping must be either 'boxes' or 'patches', got "
            + str(track_mapping)
        )



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
import bitarray as ba
import numpy as np
from os.path import abspath

from dyada.descriptor import (
    RefinementDescriptor,
    validate_descriptor,
    find_uniqueness_violations,
)
from dyada.drawing import discretization_to_2d_ascii
from dyada.refinement import (
    Discretization,
    PlannedAdaptiveRefinement,
    apply_single_refinement,
    get_defaultdict_for_markers,
    normalize_discretization,
)
from dyada.linearization import MortonOrderLinearization

from test_refinement import helper_check_mapping


def test_refine_2d_only_leaves():
    desc_initial = RefinementDescriptor(2, [1, 2])
    discretization = Discretization(MortonOrderLinearization(), desc_initial)
    p = PlannedAdaptiveRefinement(discretization)
    refinements_to_apply = [
        (0, np.array([0, 1])),
        (0, np.array([0, 1])),
        (1, np.array([1, 0])),
        (3, np.array([3, 2])),
        (2, np.array([1, 1])),
        (4, np.array([1, 1])),
        (6, np.array([0, 1])),
    ]

    for box_index, refinement in refinements_to_apply:
        p.plan_refinement(box_index, refinement)

    new_discretization, _ = p.apply_refinements()
    assert validate_descriptor(new_discretization.descriptor)


def test_refine_3d_only_leaves():
    descriptor = RefinementDescriptor(3, [1, 0, 1])
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "101")
    p.plan_refinement(1, "001")
    p.plan_refinement(2, "010")
    new_discretization, _ = p.apply_refinements()
    new_descriptor = new_discretization.descriptor
    assert new_descriptor.get_num_boxes() == 9
    assert validate_descriptor(new_descriptor)


def test_refine_simplest_not_only_leaves():
    discretization = Discretization(
        MortonOrderLinearization(), RefinementDescriptor(2, [1, 0])
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "01")

    assert len(p._planned_refinements) == 1
    assert p._planned_refinements[0][0] == 1
    assert all(p._planned_refinements[0][-1] == [0, 1])

    # don't do this at home -- call p.apply_refinements() directly
    p.populate_queue()
    assert len(p._planned_refinements) == 0
    assert len(p._markers) == 1 and all(p._markers[1] == [0, 1])
    assert p._upward_queue.queue == [(-1, 1)]

    p.upwards_sweep()
    assert len(p._markers) == 1 and all(p._markers[1] == [0, 1])
    assert p._upward_queue.empty()

    new_discretization, _ = p.create_new_discretization()
    assert new_discretization.descriptor._data == ba.bitarray("1001000000")
    assert validate_descriptor(new_discretization.descriptor)

    p = PlannedAdaptiveRefinement(new_discretization)
    p.plan_refinement(2, "01")

    new_discretization_2, _ = p.apply_refinements()
    assert new_discretization_2.descriptor == RefinementDescriptor(2, [1, 1])
    assert validate_descriptor(new_discretization_2.descriptor)


def test_refine_simplest_grandchild_split():
    discretization = Discretization(
        MortonOrderLinearization(), RefinementDescriptor(2, [1, 0])
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "01")
    p.plan_refinement(1, "10")

    assert len(p._planned_refinements) == 2
    assert p._planned_refinements[0][0] == 1
    assert all(p._planned_refinements[0][-1] == [0, 1])
    assert p._planned_refinements[1][0] == 2
    assert all(p._planned_refinements[1][-1] == [1, 0])

    # don't do this at home -- call p.apply_refinements() directly
    p.populate_queue()
    assert len(p._planned_refinements) == 0
    assert (
        len(p._markers) == 2
        and all(p._markers[1] == [0, 1])
        and all(p._markers[2] == [1, 0])
    )
    assert p._upward_queue.queue == [(-1, 1), (-1, 2)]

    p.upwards_sweep()
    assert (
        len(p._markers) == 2
        and all(p._markers[1] == [0, 1])
        and all(p._markers[2] == [1, 0])
    )
    assert p._upward_queue.empty()

    new_discretization, index_mapping = p.create_new_discretization(
        track_mapping="boxes"
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(8, 4))
        == """\
_________
|   | | |
|___| | |
|   | | |
|___|_|_|"""
    )
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("10010000100000")
    assert validate_descriptor(new_descriptor)
    helper_check_mapping(index_mapping, discretization, new_discretization)

    # plot_tree_tikz(new_descriptor, filename="simplest_grandchild_split_before")
    # plot_all_boxes_2d(Discretization(MortonOrderLinearization(), new_descriptor), labels="patches")
    p = PlannedAdaptiveRefinement(new_discretization)
    p.plan_refinement(2, ba.bitarray("01"))
    p.plan_refinement(3, ba.bitarray("01"))
    assert len(p._planned_refinements) == 2
    assert p._planned_refinements[0][0] == 5
    assert all(p._planned_refinements[0][-1] == [0, 1])
    assert p._planned_refinements[1][0] == 6
    assert all(p._planned_refinements[1][-1] == [0, 1])

    p.populate_queue()
    p.upwards_sweep()
    assert (
        len(p._markers) == 2
        and all(p._markers[0] == [0, 1])
        and all(p._markers[1] == [0, -1])
    )
    assert p._upward_queue.empty()
    new_discretization_2, index_mapping_2 = p.create_new_discretization(
        track_mapping="patches"
    )
    assert (
        discretization_to_2d_ascii(new_discretization_2, resolution=(8, 4))
        == """\
_________
|   | | |
|___|_|_|
|   | | |
|___|_|_|"""
    )
    new_descriptor_2 = new_discretization_2.descriptor
    assert new_descriptor_2._data == ba.bitarray("11 00 10 00 00 00 10 00 00")
    assert validate_descriptor(new_descriptor_2)
    helper_check_mapping(
        index_mapping_2,
        new_discretization,
        new_discretization_2,
        mapping_indices_are_boxes=False,
    )
    expected_index_mapping_2 = {
        0: {0},
        1: {0},
        2: {1},
        3: {5},
        4: {0, 2, 6},
        5: {0, 2, 6, 3, 7},
        6: {0, 2, 4, 6, 8},
    }
    assert index_mapping_2 == [
        expected_index_mapping_2[i] for i in range(len(expected_index_mapping_2))
    ]


def test_refine_grandchild_split():
    p = PlannedAdaptiveRefinement(
        Discretization(MortonOrderLinearization(), RefinementDescriptor(2, [1, 1]))
    )
    # prepare the initial state
    p.plan_refinement(0, ba.bitarray("01"))
    p.plan_refinement(2, ba.bitarray("10"))
    discretization, _ = p.apply_refinements()
    new_discretization, patch_mapping = apply_single_refinement(discretization, 1)
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == """\
_________________
|   |   |       |
|   |   |       |
|   |   |       |
|___|___|_______|
|___|___|       |
|___|___|       |
|       |       |
|_______|_______|"""
    )
    new_descriptor = new_discretization.descriptor
    four_branch = new_descriptor.get_branch(4, False)[0]
    assert new_descriptor.get_ancestry(four_branch) == [0, 1, 3]
    helper_check_mapping(patch_mapping, discretization, new_discretization)

    p = PlannedAdaptiveRefinement(new_discretization)

    # the actual test
    p.plan_refinement(0, ba.bitarray("10"))
    new_discretization, box_mapping = p.apply_refinements(track_mapping="boxes")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == """\
_________________
|   |   |       |
|   |   |       |
|   |   |       |
|___|___|_______|
|___|___|       |
|___|___|       |
|   |   |       |
|___|___|_______|"""
    )
    new_descriptor = new_discretization.descriptor
    assert validate_descriptor(new_descriptor)
    assert new_descriptor._data == ba.bitarray("111100000100000100000010000000")

    # test the mapping of the boxes
    former_to_now = [
        {0, 1},
        {2},
        {4},
        {3},
        {5},
        {6},
        {7},
        {8},
        {9},
    ]
    assert box_mapping == former_to_now


def test_refine_multi_grandchild_split():
    discretization = Discretization(
        MortonOrderLinearization(), RefinementDescriptor(2, [2, 0])
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(2, ba.bitarray("10"))
    p.plan_refinement(3, ba.bitarray("01"))
    discretization, _ = p.apply_refinements()

    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(2, ba.bitarray("01"))
    p.plan_refinement(3, ba.bitarray("11"))
    p.plan_refinement(3, ba.bitarray("10"))
    p.plan_refinement(3, ba.bitarray("10"))

    new_discretization, index_mapping = p.apply_refinements(track_mapping="boxes")
    assert new_discretization.descriptor._data == ba.bitarray(
        "10 10 00 00 11 10 00 10 10 10 00 00"
        "10 00 00 10 10 00 00 10 00 00 00 10"
        "00 10 10 10 00 00 10 00 00 10 10 00 00 10 00 00 00"
    )
    helper_check_mapping(index_mapping, discretization, new_discretization)


def test_refine_fully():
    for d in range(1, 6):
        for level in range(1, 2):
            descriptor = RefinementDescriptor(d, level)
            discretization = Discretization(MortonOrderLinearization(), descriptor)
            p = PlannedAdaptiveRefinement(discretization)
            for i in range(1, descriptor.get_num_boxes() - 1):
                # do octree refinement
                p.plan_refinement(i, ba.bitarray("1" * d))
            new_discretization, _ = p.apply_refinements()
            assert validate_descriptor(new_discretization.descriptor)
            # now also refine the first and last box
            p = PlannedAdaptiveRefinement(new_discretization)
            p.plan_refinement(0, ba.bitarray("1" * d))
            p.plan_refinement(
                new_discretization.descriptor.get_num_boxes() - 1, ba.bitarray("1" * d)
            )

            new_discretization, _ = p.apply_refinements()
            assert validate_descriptor(new_discretization.descriptor)
            regular_descriptor = RefinementDescriptor(d, level + 1)
            assert new_discretization.descriptor == regular_descriptor


def test_refine_2d_1():
    prependable_string = "10 01 00 00"
    descriptor = RefinementDescriptor.from_binary(
        2,
        ba.bitarray(prependable_string + "10 11 00 00 00 00 00"),
    )
    validate_descriptor(descriptor)
    correct_descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 00 10 10 00 00 00 00 10 10 00 00 00")
    )
    validate_descriptor(correct_descriptor)

    num_boxes_before = descriptor.get_num_boxes()
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(16, 8))
        == """\
_________________
|       | | |   |
|       | | |   |
|       | | |   |
|_______|_|_|   |
|       | | |   |
|       | | |   |
|       | | |   |
|_______|_|_|___|"""
    )

    new_discretization, index_mapping = apply_single_refinement(
        discretization, len(discretization) - 1, ba.bitarray("01"), "patches"
    )
    new_discretization_, box_mapping = apply_single_refinement(
        discretization, len(discretization) - 1, ba.bitarray("01")
    )
    assert new_discretization == new_discretization_
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == """\
_________________
|       | | |   |
|       | | |   |
|       | | |   |
|_______|_|_|___|
|       | | |   |
|       | | |   |
|       | | |   |
|_______|_|_|___|"""
    )

    new_descriptor = new_discretization.descriptor
    validate_descriptor(new_descriptor)
    assert new_descriptor.get_num_boxes() == num_boxes_before + 1
    assert new_descriptor == correct_descriptor
    expected_box_mapping = {
        0: {0},
        1: {4},
        2: {1},
        3: {2},
        4: {5},
        5: {6},
        6: {3, 7},
    }
    assert box_mapping == list(expected_box_mapping.values())
    expected_index_mapping = {
        0: {0},
        1: {0},
        2: {1},
        3: {7},
        4: {0, 2, 8},
        5: {0, 2, 3, 8, 9},
        6: {4},
        7: {5},
        8: {10},
        9: {11},
        10: {0, 2, 8, 6, 12},
    }
    assert index_mapping == list(expected_index_mapping.values())
    helper_check_mapping(box_mapping, discretization, new_discretization)
    helper_check_mapping(index_mapping, discretization, new_discretization, False)


def test_refine_2d_2():
    descriptor = RefinementDescriptor.from_binary(
        2,
        ba.bitarray("11 00 00 10 00 00 10 00 11 00 00 00 00"),
    )
    validate_descriptor(descriptor)
    num_boxes_before = descriptor.get_num_boxes()
    assert num_boxes_before == 9
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(16, 8))
        == """\
_________________
|   |   |   | | |
|   |   |   |_|_|
|   |   |   | | |
|___|___|___|_|_|
|       |       |
|       |       |
|       |       |
|_______|_______|"""
    )

    p = PlannedAdaptiveRefinement(discretization)
    refinements = [
        (3, ba.bitarray("11")),
        (4, ba.bitarray("01")),
    ]
    for box, refinement in refinements:
        p.plan_refinement(box, refinement)
    assert len(p._planned_refinements) == 2
    assert p._planned_refinements[0][0] == 5
    assert all(p._planned_refinements[0][-1] == [1, 1])
    assert p._planned_refinements[1][0] == 7
    assert all(p._planned_refinements[1][-1] == [0, 1])
    p.populate_queue()
    p.upwards_sweep()
    assert (
        len(p._markers) == 3
        and all(p._markers[5] == [1, 1])
        and all(p._markers[6] == [0, 1])
        and all(p._markers[8] == [0, -1])
    )
    assert p._upward_queue.empty()
    new_discretization, index_mapping = p.create_new_discretization(
        track_mapping="boxes"
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == """\
_________________
|   | | |   | | |
|   |_|_|___|_|_|
|   | | |   | | |
|___|_|_|___|_|_|
|       |       |
|       |       |
|       |       |
|_______|_______|"""
    )
    validate_descriptor(new_discretization.descriptor)

    descriptor_expected = RefinementDescriptor.from_binary(
        2,
        ba.bitarray("11 00 00 10 00 11 00 00 00 00 11 00 10 00 00 00 10 00 00"),
    )
    assert new_discretization.descriptor == descriptor_expected

    helper_check_mapping(index_mapping, discretization, new_discretization)


def test_refine_2d_3():
    """Masado's example"""
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("10 01 00 00 10 01 00 00 00")
    )
    assert find_uniqueness_violations(descriptor) == []

    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(2, "01")
    p.plan_refinement(4, "02")
    non_normalized_discretization, _ = p.apply_refinements()
    ascii_before_and_after = """\
_________
|   | |_|
|___|_|_|
|   |_|_|
|___|_|_|"""
    assert (
        discretization_to_2d_ascii(non_normalized_discretization, resolution=(8, 4))
        == ascii_before_and_after
    )
    non_normalized_descriptor = non_normalized_discretization.descriptor
    assert non_normalized_descriptor._data == ba.bitarray(
        "11 00 10 01 00 00 01 00 00 00 10 00 01 00 00"
    )
    assert find_uniqueness_violations(non_normalized_descriptor) == [{2, 3, 6}]
    p = PlannedAdaptiveRefinement(non_normalized_discretization)

    p._markers = get_defaultdict_for_markers(descriptor.get_num_dimensions())
    p._markers[2] = np.array([0, 1], dtype=np.int8)
    p._markers[3] = np.array([0, -1], dtype=np.int8)
    p._markers[6] = np.array([0, -1], dtype=np.int8)

    new_discretization, _ = p.create_new_discretization(track_mapping="patches")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(8, 4))
        == ascii_before_and_after
    )
    new_descriptor = new_discretization.descriptor
    assert new_descriptor._data == ba.bitarray("11 00 11 00 00 00 00 00 10 00 01 00 00")
    assert find_uniqueness_violations(new_descriptor) == []
    normalized_discretization, mapping, num_rounds = normalize_discretization(
        non_normalized_discretization
    )
    assert num_rounds == 1
    assert normalized_discretization.descriptor == new_descriptor
    assert mapping == [
        {0},
        {1},
        {2},
        {2},
        {3},
        {5},
        {2},
        {4},
        {6},
        {7},
        {8},
        {9},
        {10},
        {11},
        {12},
    ]


def test_refine_2d_4():
    "Masado's more complex example"
    non_normalized_descriptor = RefinementDescriptor.from_binary(
        2,
        ba.bitarray(
            "10 10 01 01 11 00 00 00 00 00 00 \
            10 01 01 01 00 00 00 00 01 01 01 00 00 00 00 01 00 00"
        ),
    )
    assert find_uniqueness_violations(non_normalized_descriptor) == [{11, 12, 19}]
    ascii_before_and_after = """\
_________________
|   | | |       |
|   | | |       |
|   | | |       |
|___|_|_|_______|
|   | | |       |
|___|_|_|       |
|_|_|_|_|       |
|_|_|_|_|_______|"""

    non_normalized_discretization = Discretization(
        MortonOrderLinearization(), non_normalized_descriptor
    )
    new_discretization, mapping, num_rounds = normalize_discretization(
        non_normalized_discretization,
        track_mapping="patches",
        max_normalization_rounds=1,
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == ascii_before_and_after
    )
    assert num_rounds == 1
    assert new_discretization.descriptor == RefinementDescriptor.from_binary(
        2,
        ba.bitarray(
            "10 10 01 01 11 00 00 00 00 00 00 11 01 01 00 00 00 01 01 00 00 00 00 00 01 00 00"
        ),
    )
    assert find_uniqueness_violations(new_discretization.descriptor) == [{1, 2, 11}]
    new_discretization, mapping, num_rounds = normalize_discretization(
        non_normalized_discretization,
        track_mapping="boxes",
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == ascii_before_and_after
    )
    assert num_rounds == 5
    assert new_discretization.descriptor == RefinementDescriptor.from_binary(
        2,
        ba.bitarray(
            "11 11 11 00 00 00 00 11 00 00 00 00 00 10 00 00 00 10 00 10 00 00 00"
        ),
    )
    assert find_uniqueness_violations(new_discretization.descriptor) == []
    assert mapping == [
        {0},
        {1},
        {2},
        {3},
        {8},
        {12},
        {4},
        {6},
        {9},
        {13},
        {5},
        {7},
        {10},
        {14},
        {11},
        {15},
    ]
    _, patch_mapping, num_rounds = normalize_discretization(
        non_normalized_discretization,
        track_mapping="patches",
    )
    assert num_rounds == 5
    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 17},
        2: {0, 1, 2, 17},
        3: {1, 2},
        4: {2},
        5: {3},
        6: {4},
        7: {5},
        8: {6},
        9: {12},
        10: {18},
        11: {0, 1, 7, 13, 17, 19},
        12: {0, 1, 7, 13, 17, 19},
        13: {1, 7, 13},
        14: {7},
        15: {8},
        16: {10},
        17: {14},
        18: {20},
        19: {0, 1, 7, 13, 17, 19},
        20: {1, 7, 13},
        21: {7},
        22: {9},
        23: {11},
        24: {15},
        25: {21},
        26: {0},
        27: {16},
        28: {22},
    }
    assert patch_mapping == list(expected_patch_mapping.values())


def test_refine_2d_5():
    """Test found through randomized testing, it used to return incomplete mapping"""
    descriptor = RefinementDescriptor.from_binary(
        2,
        ba.bitarray("10 00 01 01 00 00 01 00 00"),
    )
    assert validate_descriptor(descriptor)
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == """\
_________
|   |___|
|   |___|
|   |___|
|___|___|"""
    )

    refinement = ba.bitarray("01")
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, refinement)
    new_discretization, index_mapping = p.apply_refinements(track_mapping="patches")

    assert validate_descriptor(new_discretization.descriptor)
    helper_check_mapping(
        index_mapping,
        discretization,
        new_discretization,
        mapping_indices_are_boxes=False,
        tested_refinement=[refinement],
    )

    assert new_discretization.descriptor._data == ba.bitarray(
        "11 00 01 00 00 00 01 00 00"
    )
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(8, 4))
        == """\
_________
|   |___|
|___|___|
|   |___|
|___|___|"""
    )
    expected_mapping = {
        0: {0},
        1: {0, 1, 5},
        2: {0, 2, 6},
        3: {2},
        4: {3},
        5: {4},
        6: {6},
        7: {7},
        8: {8},
    }
    assert index_mapping == [expected_mapping[i] for i in range(len(expected_mapping))]


def test_refine_2d_6():
    discretization = Discretization(
        MortonOrderLinearization(),
        RefinementDescriptor.from_binary(
            2,
            ba.bitarray("01 00 10 10 00 00 10 00 00"),
        ),
    )
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == """\
_________
| | | | |
|_|_|_|_|
|       |
|_______|"""
    )
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_refinement(0, "10")
    p.plan_refinement(1, "10")
    p.plan_refinement(2, "10")
    new_discretization, patch_mapping = p.apply_refinements(track_mapping="patches")
    assert (
        discretization_to_2d_ascii(new_discretization, resolution=(16, 8))
        == """\
_________________
| | | | |   |   |
| | | | |   |   |
| | | | |   |   |
|_|_|_|_|___|___|
|       |       |
|       |       |
|       |       |
|_______|_______|"""
    )
    validate_descriptor(new_discretization.descriptor)
    expected_patch_mapping = {
        0: {0},
        1: {0, 1, 2},
        2: {0, 3, 10},
        3: {3, 4, 7},
        4: {4, 5, 6},
        5: {7, 8, 9},
        6: {10},
        7: {11},
        8: {12},
    }
    assert patch_mapping == [
        expected_patch_mapping[i] for i in range(len(expected_patch_mapping))
    ]
    helper_check_mapping(
        patch_mapping,
        discretization,
        new_discretization,
        mapping_indices_are_boxes=False,
    )


if __name__ == "__main__":
    here = abspath(__file__)
    pytest.main([here, "-s"])



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from dyada.coordinates import (
    CoordinateInterval,
    get_coordinates_from_level_index,
    interval_from_sequences,
    DyadaTooFineError,
)
from dyada.descriptor import RefinementDescriptor
from dyada.discretization import (
    Discretization,
    coordinates_from_box_index,
    SliceDictInDimension,
    get_level_index_from_linear_index,
)
from dyada.drawing import (  # TODO move to drawing once ruff allows finer control
    discretization_to_2d_ascii,
    plot_all_boxes_2d,
    plot_all_boxes_3d,
    get_figure_2d_matplotlib,
    get_figure_3d_matplotlib,
)
from dyada.drawing_opengl_obj import (
    plot_boxes_3d_pyopengl,
    export_boxes_3d_to_obj,
)
from dyada.drawing_tikz import (
    plot_tree_tikz,
    plot_descriptor_tikz,
    plot_boxes_2d_tikz,
    plot_boxes_3d_tikz,
    plot_location_stack_tikz,
)
from dyada.linearization import MortonOrderLinearization
from dyada.refinement import PlannedAdaptiveRefinement, apply_single_refinement

assert sys.version_info >= (3, 10), "Use Python 3.10 or newer"



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
import bitarray as ba
from dyada.descriptor import (
    RefinementDescriptor,
    validate_descriptor,
    find_uniqueness_violations,
)
from dyada.refinement import (
    Discretization,
    PlannedAdaptiveRefinement,
)
from dyada.drawing import discretization_to_2d_ascii
from dyada.linearization import MortonOrderLinearization
from test_refinement import helper_check_mapping


def test_downsplit_uncanonicalize_2d():
    descriptor = RefinementDescriptor.from_binary(2, ba.bitarray("11 00 00 00 00"))
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    ascii_before_and_after = """\
_____
|_|_|
|_|_|"""
    assert str(discretization) == ascii_before_and_after
    # Keep downsplit markers exactly as requested
    down_plan = PlannedAdaptiveRefinement(discretization)
    down_plan.plan_downsplit(0, ba.bitarray("10"))
    down_discretization, _ = down_plan.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert str(down_discretization) == ascii_before_and_after
    down_descriptor = down_discretization.descriptor
    assert down_descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 00 00 10 00 00")
    )
    assert find_uniqueness_violations(down_descriptor) == [{0, 1, 4}]


def test_downsplit_child():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 10 00 00 00 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    ascii_before_and_after = """\
_________
|___|___|
|_|_|___|"""
    assert str(discretization) == ascii_before_and_after

    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("10"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    # descriptor must have changed, e.g. starting with 01
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 10 00 00 00 10 00 00")
    )
    assert str(down_discretization) == ascii_before_and_after
    assert patch_mapping == [
        {0, 1, 6},
        {2},
        {3},
        {4},
        {5},
        {7},
        {8},
    ]
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )

    # again in the other direction
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("01"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("10 01 10 00 00 00 01 00 00")
    )
    assert str(down_discretization) == ascii_before_and_after
    assert patch_mapping == [
        {0, 1, 6},
        {2},
        {3},
        {4},
        {7},
        {5},
        {8},
    ]
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )


def test_downsplit_child_mixed_root_children_2d():
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 01 00 00 10 00 00 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    ascii_before_and_after = """\
_________
|   |   |
|___|___|
|___| | |
|___|_|_|"""
    assert str(discretization) == ascii_before_and_after

    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("10"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 01 00 00 10 00 00 10 00 00")
    )
    assert str(down_discretization) == ascii_before_and_after
    assert patch_mapping == [
        {0, 1, 8},
        {2},
        {3},
        {4},
        {5},
        {6},
        {7},
        {9},
        {10},
    ]
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )


def test_downsplit_root_3d():
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("111 000 000 000 000 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("100"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        3, ba.bitarray("011 100 000 000 100 000 000 100 000 000 100 000 000")
    )
    assert patch_mapping == [
        {0, 1, 4, 7, 10},
        {2},
        {3},
        {5},
        {6},
        {8},
        {9},
        {11},
        {12},
    ]
    assert descriptor.get_num_boxes() == down_discretization.descriptor.get_num_boxes()
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )


def test_downsplit_nonroot_3d_mixed_children():
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("111 110 010 000 000 000 000 000 000 000 000 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(1, ba.bitarray("100"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "111 010 100 010 000 000 000 100 000 000 000 000 000 000 000 000 000"
        ),
    )
    assert patch_mapping == [
        {0},
        {1, 2, 7},
        {3},
        {4},
        {5},
        {6},
        {8},
        {9},
        {10},
        {11},
        {12},
        {13},
        {14},
        {15},
        {16},
    ]
    assert descriptor.get_num_boxes() == down_discretization.descriptor.get_num_boxes()
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )


def test_downsplit_2dims_simultaneously_3d():
    # Split two dimensions at once from a fully-refined 3D root with leaf children.
    # When both grouped children can absorb the split, we merge directly
    # into those children instead of inserting intermediate nodes.
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("111 000 000 000 000 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("110"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        3, ba.bitarray("001 110 000 000 000 000 110 000 000 000 000")
    )
    assert patch_mapping == [
        {0, 1, 6},
        {2},
        {3},
        {4},
        {5},
        {7},
        {8},
        {9},
        {10},
    ]
    assert descriptor.get_num_boxes() == down_discretization.descriptor.get_num_boxes()
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )


def test_downsplit_merged_plans_3d():
    # Two separate plan_downsplit calls on the same node are OR-merged before
    # application and must produce the same result as a single combined call.
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("111 000 000 000 000 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("100"))
    p.plan_downsplit(0, ba.bitarray("010"))
    down_discretization, patch_mapping = p.apply_refinements(
        track_mapping="patches",
        sweep_mode="as_planned",
    )
    assert down_discretization.descriptor == RefinementDescriptor.from_binary(
        3, ba.bitarray("001 110 000 000 000 000 110 000 000 000 000")
    )
    assert patch_mapping == [
        {0, 1, 6},
        {2},
        {3},
        {4},
        {5},
        {7},
        {8},
        {9},
        {10},
    ]
    validate_descriptor(down_discretization.descriptor)
    helper_check_mapping(
        patch_mapping,
        discretization,
        down_discretization,
        mapping_indices_are_boxes=False,
    )


def test_downsplit_boxes_mapping_3d():
    # Downsplit only restructures intermediate nodes; it never splits or merges
    # leaf boxes, so the boxes mapping must be the identity.
    descriptor = RefinementDescriptor.from_binary(
        3, ba.bitarray("111 000 000 000 000 000 000 000 000")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("100"))
    down_discretization, boxes_mapping = p.apply_refinements(
        track_mapping="boxes",
        sweep_mode="as_planned",
    )
    assert boxes_mapping == [{i} for i in range(8)]


def test_downsplit_only_dim_not_implemented():
    # Downsplitting the sole remaining refined dimension would leave the node with no
    # refinement at all; this path is not yet supported.
    descriptor = RefinementDescriptor.from_binary(2, ba.bitarray("10 00 00"))
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("10"))
    with pytest.raises(ValueError):
        p.apply_refinements(sweep_mode="as_planned")


def test_downsplit_nested_2d():
    """Downsplit both root and its child in the same pass."""
    descriptor = RefinementDescriptor.from_binary(
        2, ba.bitarray("11 11 00 00 00 00 00 00 00")
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    assert descriptor.get_num_boxes() == 7
    ascii_before_and_after = """\
_________
|   |   |
|___|___|
|_|_|   |
|_|_|___|"""
    assert (
        discretization_to_2d_ascii(discretization, resolution=(8, 4))
        == ascii_before_and_after
    )

    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("10"))
    p.plan_downsplit(1, ba.bitarray("10"))
    new_disc, mapping = p.apply_refinements(
        track_mapping="patches", sweep_mode="as_planned"
    )
    new_desc = new_disc.descriptor

    # Leaf count and spatial discretization must be preserved
    assert new_desc.get_num_boxes() == 7
    assert validate_descriptor(new_desc)
    assert (
        discretization_to_2d_ascii(new_disc, resolution=(8, 4))
        == ascii_before_and_after
    )
    assert new_desc == RefinementDescriptor.from_binary(
        2, ba.bitarray("01 10 01 10 00 00 10 00 00 00 10 00 00")
    )


def test_downsplit_nested_3d():
    descriptor = RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "111 111 000 000 000 000 000 000 000 000 000 000 000 000 000 000 000"
        ),
    )
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    num_boxes = descriptor.get_num_boxes()

    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("100"))
    p.plan_downsplit(1, ba.bitarray("100"))
    new_disc, _ = p.apply_refinements(track_mapping="patches", sweep_mode="as_planned")
    new_desc = new_disc.descriptor

    assert new_desc.get_num_boxes() == num_boxes
    assert validate_descriptor(new_desc)
    # Root becomes 011, each child gets 100 intermediate with nested downsplit
    assert new_desc == RefinementDescriptor.from_binary(
        3,
        ba.bitarray(
            "011"
            " 100 011 100 000 000 100 000 000 100 000 000 100 000 000 000"
            " 100 000 000"
            " 100 000 000"
            " 100 000 000"
        ),
    )


def test_downsplit_invalid_track_mapping():
    descriptor = RefinementDescriptor.from_binary(2, ba.bitarray("11 00 00 00 00"))
    discretization = Discretization(MortonOrderLinearization(), descriptor)
    p = PlannedAdaptiveRefinement(discretization)
    p.plan_downsplit(0, ba.bitarray("10"))
    with pytest.raises(ValueError, match="track_mapping"):
        p.apply_refinements(track_mapping="invalid", sweep_mode="as_planned")



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

import warnings

try:
    from cmap import Colormap
except ImportError:
    warnings.warn("cmap not found, some plotting functions will not work")
from itertools import product
from string import ascii_uppercase
from typing import Sequence, Union, Mapping

from dyada.coordinates import (
    Coordinate,
    CoordinateInterval,
)
from dyada.discretization import (
    Discretization,
)
from dyada.structure import depends_on_optional


def labels_from_discretization(
    discretization: Discretization, labels: Union[None, str, Sequence[str]]
):
    if labels == "patches":
        labels = []
        for i in range(len(discretization._descriptor)):
            if discretization._descriptor.is_box(i):
                labels.append(str(i))
    elif labels == "boxes":
        labels = [str(i) for i in range(len(discretization))]

    assert labels is None or len(labels) == discretization._descriptor.get_num_boxes()
    return labels


def cuboid_from_interval(
    interval: CoordinateInterval, projection: Sequence[int]
) -> tuple:
    lower = interval[0][projection]
    upper = interval[1][projection]
    # create vertices from lower and upper bounds
    vertices = list(product(*zip(lower, upper)))
    # define rectangular faces
    faces = (
        (0, 1, 3, 2),
        (4, 6, 7, 5),
        (0, 2, 6, 4),
        (1, 5, 7, 3),
        (0, 4, 5, 1),
        (2, 3, 7, 6),
    )
    edges = (
        *((0, 1), (2, 3), (4, 5), (6, 7)),
        *((0, 2), (1, 3), (4, 6), (5, 7)),
        *((0, 4), (1, 5), (2, 6), (3, 7)),
    )
    return vertices, faces, edges


def side_corners_generator(
    lower_cube_corner: Coordinate, upper_cube_corner: Coordinate
):
    # iterate the six sides of the cuboid
    # by always selecting four corners that have one coordinate in common
    corners = list(product(*zip(lower_cube_corner, upper_cube_corner)))
    for bound in [lower_cube_corner, upper_cube_corner]:
        for i, b in enumerate(bound):
            side_corners = list(filter(lambda c: c[i] == b, corners))
            assert len(side_corners) == 4
            # correct the order
            yield side_corners[0], side_corners[1], side_corners[3], side_corners[2]


def overwrite_char_in_canvas(
    canvas: list[list[str]], i: int, j: int, vertical: bool
) -> list[list[str]]:
    new = "|" if vertical else "_"
    old = canvas[i][j]

    if old not in (new, " "):
        W = len(canvas[0])
        ok = not vertical or (
            j == 0 or j == W - 1 or canvas[i][j - 1] == "_" or canvas[i][j + 1] == "_"
        )

        if not ok:
            canvas[i][j] = "*"
            drawing = "\n".join("".join(r) for r in canvas)
            warnings.warn(f"Overwriting non-whitespace with {new}\n{drawing}")

    canvas[i][j] = new
    return canvas


def boxes_to_2d_ascii(
    intervals: Union[Sequence[CoordinateInterval], Mapping[CoordinateInterval, str]],
    projection: Sequence[int],
    resolution=(16, 8),
    **kwargs,
) -> str:
    """
    returns an ASCII string visualization of a 2-D omnitree discretization
    using | and _ characters for cell boundaries.

    cells: list of ((x0, x1), (y0, y1)) dyadic rectangles
    resolution: grid resolution in (width, height) (must be power-of-two compatible)
    """

    def to_idx(x, dim):
        # unit coordinates -> grid indices
        # dim=0: horizontal direction, dim=1: vertical direction
        idx = int(round(x * resolution[dim]))
        if abs(idx - (x * resolution[dim])) > 1e-8:
            warnings.warn(
                f"Coordinate {x} in not aligned with resolution {resolution[dim]}[{dim}]"
            )
        return idx

    # canvas with boundary slots
    W = resolution[0] + 1
    H = resolution[1] + 1
    canvas = [[" " for _ in range(W)] for _ in range(H)]

    # draw cell boundaries
    for interval in intervals:
        lower = interval[0][projection]
        upper = interval[1][projection]
        ix0 = to_idx(lower[0], 0)
        ix1 = to_idx(upper[0], 0)
        iy0 = to_idx(lower[1], 1)
        iy1 = to_idx(upper[1], 1)

        # Horizontal edges
        for x in range(ix0, ix1 + 1):
            canvas = overwrite_char_in_canvas(canvas, H - iy0 - 1, x, False)
            canvas = overwrite_char_in_canvas(canvas, H - iy1 - 1, x, False)

        # Vertical edges
        for y in range(iy0, iy1):
            canvas = overwrite_char_in_canvas(canvas, H - y - 1, ix0, True)
            canvas = overwrite_char_in_canvas(canvas, H - y - 1, ix1, True)

    return "\n".join("".join(row) for row in canvas)


@depends_on_optional("cmap")
def get_colors(num_colors: int, colormap_name="CET_R3"):
    cm = Colormap(colormap_name)
    for leaf in range(num_colors):
        colormapped = cm(leaf * 1.0 / num_colors)
        yield colormapped  # RGB values in [0, 1] range


def get_colors_byte(num_colors: int, colormap_name="CET_R3"):
    for color in get_colors(num_colors, colormap_name):
        color = [int(255 * c) for c in color]  # convert to [0, 255] range
        yield color


def letter_counter(length):
    """Generate up to `length` Excel-style letter labels (A, B, ..., Z, AA, AB, ...)."""
    count = 0
    size = 1
    while count < length:
        for combo in product(ascii_uppercase, repeat=size):
            yield "".join(combo)
            count += 1
            if count >= length:
                break
        size += 1



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections import defaultdict
import numpy as np
import numpy.typing as npt
from types import MappingProxyType
from typing import TypeAlias

MarkerType: TypeAlias = npt.NDArray[np.int8]
MarkersType: TypeAlias = defaultdict[int, npt.NDArray[np.int8]]
MarkersMapProxyType: TypeAlias = MappingProxyType[int, npt.NDArray[np.int8]]


def filter_markers_by_min_index(
    markers: MarkersType | MarkersMapProxyType, min_index: int
) -> MarkersMapProxyType:
    # filter the markers to the current interval
    filtered_markers = {k: v for k, v in markers.items() if k >= min_index}
    return MarkersMapProxyType(filtered_markers)


def get_next_largest_markered_index(
    markers: MarkersType | MarkersMapProxyType, min_index: int
) -> int:
    return min(
        filter_markers_by_min_index(markers, min_index).keys(),
        default=-1,
    )


def get_defaultdict_for_markers(
    num_dimensions: int,
) -> MarkersType:
    def get_d_zeros_as_array():
        return np.zeros(num_dimensions, dtype=np.int8)

    return MarkersType(get_d_zeros_as_array)



# SPDX-FileCopyrightText: 2025 Theresa Pollinger
#
# SPDX-License-Identifier: GPL-3.0-or-later

from copy import deepcopy
from functools import lru_cache, wraps
from importlib.util import find_spec


def module_is_available(module_name: str) -> bool:
    return find_spec(module_name) is not None


def depends_on_optional(module_name: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not module_is_available(module_name):
                raise ImportError(
                    f"Optional dependency {module_name} not found ({func.__name__})."
                )
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator


# cf. https://stackoverflow.com/a/54909677/7272382
def copying_lru_cache():  # TODO add args for lru_cache as needed
    def decorator(func):
        cached_func = lru_cache()(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return deepcopy(cached_func(*args, **kwargs))

        return wrapper

    return decorator
